from pathlib import Path
import json
import yaml
from typing import Dict, List, Any
from rich.console import Console

# Custom modules
from optrade.data.universe import Universe
from optrade.dev.utils.ablations import generate_ablation_combinations

SCRIPT_DIR = Path(__file__).resolve().parent
UNIVERSE_DOWNLOAD_PARAMETERS = [
    "contracts.stride",
    "contracts.interval_min",
    "contracts.right",
    "contracts.target_tte",
    "contracts.tte_tolerance",
    "contracts.moneyness",
    "data.train_split",
    "data.val_split",
    "contracts.strike_band",
    "contracts.volatility_type",
    "contracts.volatility_scaled",
    "contracts.volatility_scalar",
]


def filter_ablation_config(
    ablation_config: Dict[str, List[Any]], prefixes: List[str]
) -> Dict[str, List[Any]]:
    """
    Filter ablation configuration by prefixes and group them
    Args:
        ablation_config: The full ablation configuration
        prefixes: List of prefixes to filter by (e.g. ["data", "contracts"])
    Returns:
        Dictionary with filtered and grouped ablation parameters
    """
    # ctx = Console()
    filtered_config = ablation_config.copy()
    keys_to_delete = []

    for key in filtered_config:
        # Check if the key starts with any of the prefixes
        if not any(key.startswith(prefix) for prefix in prefixes):
            keys_to_delete.append(key)

    # Delete keys outside the loop to avoid modification during iteration
    for key in keys_to_delete:
        # ctx.log(f"Removing {key} from filtered config")
        del filtered_config[key]

    return filtered_config


def run_universe(parent_id: str, download: bool = False, filter: bool = False) -> None:
    """Set up universe, filter it, and optionally download data with ablation parameters

    Args:
        parent_id: The ID of the parent experiment
        download: Whether to download data or not
        filter: Whether to filter the universe or not

    Returns:
        None
    """

    ctx = Console()

    # Set up paths
    parent_path = (
        SCRIPT_DIR.parents[0] / "jobs" / "parent" / parent_id
    )
    universe_path = parent_path / "universe.yaml"

    # Load universe configuration
    with open(universe_path, "r") as file:
        universe_config = yaml.safe_load(file)

    # Initialize universe
    universe = Universe(
        start_date=universe_config["start_date"],
        end_date=universe_config["end_date"],
        sp_500=universe_config.get("sp_500", False),
        nasdaq_100=universe_config.get("nasdaq_100", False),
        dow_jones=universe_config.get("dow_jones", False),
        candidate_roots=universe_config.get("candidate_roots", None),
        volatility=universe_config.get("volatility", None),
        pe_ratio=universe_config.get("pe_ratio", None),
        debt_to_equity=universe_config.get("debt_to_equity", None),
        beta=universe_config.get("beta", None),
        market_cap=universe_config.get("market_cap", None),
        sector=universe_config.get("sector", None),
        industry=universe_config.get("industry", None),
        dividend_yield=universe_config.get("dividend_yield", None),
        earnings_volatility=universe_config.get("earnings_volatility", None),
        market_beta=universe_config.get("market_beta", None),
        size_beta=universe_config.get("size_beta", None),
        value_beta=universe_config.get("value_beta", None),
        profitability_beta=universe_config.get("profitability_beta", None),
        investment_beta=universe_config.get("investment_beta", None),
        momentum_beta=universe_config.get("momentum_beta", None),
        all_metrics=universe_config.get("all_metrics", False),
        save_dir=universe_config.get("save_dir", None),
        verbose=universe_config.get("verbose", False),
        dev_mode=universe_config.get("dev_mode", False),
    )

    # Set up universe
    ctx.log("Setting candidate roots...")
    universe.set_roots()
    ctx.log("Getting market metrics...")
    universe.get_market_metrics()
    ctx.log("Getting factor exposures...")
    universe.get_factor_exposures()
    ctx.log(f"Roots: {universe.roots}")

    # Filter the universe
    if filter:
        universe.filter()

    # Save market metrics
    market_metrics_path = parent_path / "market_metrics.json"
    ctx.log(f"Saving market metrics to {market_metrics_path}")
    with open(market_metrics_path, "w") as f:
        json.dump(universe.market_metrics, f, indent=4)

    if download:
        # Load ablation configuration
        ablation_path = parent_path / "ablation.yaml"
        with open(ablation_path, "r") as file:
            ablation_full_config = yaml.safe_load(file)["ablations"]

        ctx.log(f"Ablation full config: {ablation_full_config}")
        ctx.log(f"Keys: {list(ablation_full_config.keys())}")

        # Filter only by "contracts" and "data", as they are the only ones needed for downloading
        filtered_ablation_config = filter_ablation_config(
            ablation_full_config, UNIVERSE_DOWNLOAD_PARAMETERS
        )
        ablation_combinations = generate_ablation_combinations(filtered_ablation_config)

        for i, ablation in enumerate(ablation_combinations):
            ctx.log(f"Downloading data combination {i+1}/{len(ablation_combinations)}")

            # Download data with these parameters
            universe.download(
                contract_stride=ablation.get("contracts.stride"),
                interval_min=ablation.get("contracts.interval_min"),
                right=ablation.get("contracts.right"),
                target_tte=ablation.get("contracts.target_tte"),
                tte_tolerance=ablation.get("contracts.tte_tolerance"),
                moneyness=ablation.get("contracts.moneyness"),
                train_split=ablation.get("data.train_split"),
                val_split=ablation.get("data.val_split"),
                strike_band=ablation.get("contracts.strike_band", 0.05),
                volatility_type=ablation.get("contracts.volatility_type", "period"),
                volatility_scaled=ablation.get("contracts.volatility_scaled", False),
                volatility_scalar=ablation.get("contracts.volatility_scalar", 1.0),
            )

        ctx.log("Download complete!")


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("parent_id", type=str, help="Parent experiment ID")
    parser.add_argument("--d", action="store_true", help="Download data")
    parser.add_argument("--f", action="store_true", help="Filter universe")

    args = parser.parse_args()

    run_universe(
        parent_id=args.parent_id,
        download=args.d,
        filter=args.f,
    )



from pathlib import Path
import itertools
import yaml
from typing import Dict, List, Any, Tuple
from rich.console import Console
import argparse

from optrade.config.config import load_config
from optrade.main import run_job
from optrade.dev.utils.ablations import load_ablation_config, generate_ablation_combinations

SCRIPT_DIR = Path(__file__).resolve().parent

def grid_search(job_name: str) -> None:
    # Adjust the base_path to use the absolute path
    base_path = SCRIPT_DIR.parents[0] / "jobs" # <- Might have to change due to refactor
    ablation_path = base_path / job_name / "ablation.yaml"

    # Rich
    ctx = Console()

    # Load the ablation configuration
    ablation_config = load_ablation_config(ablation_path)

    # Load args and check if 'parent' experiment is provided
    args_path = base_path / job_name / "args.yaml"
    args = load_config(args_path)

    if args.exp.parent is not None:
        with ctx.status(f"Adjusting ablation config according to parent: {args.exp.parent}..."):
            parent_path = base_path / "parent" / args.exp.parent / "ablation.yaml"
            parent_ablation_config = load_ablation_config(parent_path)

            original_ablation_config = ablation_config.copy()

            # Merge parent ablation config with current ablation config
            for key, value in parent_ablation_config.items():
                ablation_config[key] = value

    ctx.log(f"Original ablation config: {original_ablation_config}")
    ctx.log(f"Updated ablation config: {ablation_config}")

    # Generate all combinations of ablations
    ablation_combinations = generate_ablation_combinations(ablation_config)

    # Run each ablation
    for i, ablation in enumerate(ablation_combinations):
        ctx.log(f"Running ablation {i}: {ablation}")
        run_job(job_name=job_name, ablation=ablation, ablation_id=i)

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "job_name", type=str, default="forecasting/test", help="Name of the ablation"
    )
    args = parser.parse_args()
    grid_search(args.job_name)



import numpy as np
import torch
import pandas as pd
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from typing import Union, Tuple, Optional, List, Dict, Any
from sklearn.base import BaseEstimator
import torch.nn as nn

from optrade.data.forecasting import ForecastingDataset
from optrade.utils.misc import tensor_to_datetime, datetime_to_tensor
from scipy import stats
from sklearn.metrics import mean_squared_error, mean_absolute_error


class Analyzer:
    """
    Comprehensive analysis tool for model forecast performance evaluation.
    """

    def __init__(self):
        pass

    def period_visualize(
        self,
        period: str,
        period_interval: int,
        model: Union[nn.Module, BaseEstimator],
        dataset: Union[ForecastingDataset, np.ndarray],
        metrics: List[str],
        batch_size: int = 128,
        x_axis: str = "Time of Day",
        y_axis: str = "Normalized Error",
        title: Optional[str] = None,
        figsize: Tuple[int, int] = (12, 6),
        normalize: bool = False,
        use_secondary_axis: bool = False,
        dpi: int = 300,
        output_format: str = "png",
        save: bool = False,
    ) -> plt.Figure:
        """
        Visualize metrics aggregated by specific time periods across multiple days.
        Args:
            period (str): Type of period to group by. Options: "daily". Other periods not yet implemented.
            period_interval (int): If period="daily", period_interval represents the number of minutes to group by.
            model: Model object (PyTorch or scikit-learn)
            dataset: ForecastingDataset or numpy array of time series data
            metrics (List[str]): List of metrics to calculate ("mse", "mae", "rmse", "mape", "r2")
            batch_size (int): Batch size for DataLoader
            x_axis (str): Label for x-axis
            y_axis (str): Label for y-axis
            title (Optional[str]): Plot title
            figsize (Tuple[int, int]): Figure size (width, height)
            normalize (bool): Whether to normalize metrics to [0,1] range for comparison
            use_secondary_axis (bool): Use a secondary y-axis for the second metric
            dpi (int): Dots per inch for image resolution
            output_format (str): Output format for saving the image
            save (bool): Save the plot to disk
        Returns:
            plt.Figure: The matplotlib figure object
        """
        # Get predictions and targets
        if isinstance(model, nn.Module):
            preds, targets, targets_dt = self.get_torch_preds(
                model, dataset, batch_size
            )
            # Flatten and convert targets_dt to pandas datetime
            preds = preds.reshape(-1)
            targets = targets.reshape(-1)
            targets_dt = pd.to_datetime(targets_dt.reshape(-1))
        else:
            raise NotImplementedError(
                "Only PyTorch models are supported for now. Scikit-learn will be implemented later."
            )

        # Extract the time component based on period_type
        if period == "daily":
            # Set intervals from 9:30 AM to 4:00 PM (trading hours). Stride by period_interval
            intervals = pd.date_range(
                "09:30", "16:00", freq=f"{period_interval}min"
            ).time

            # Create a dictionary to store metrics for each interval
            results = {}
            interval_labels = []

            # Iterate through each interval
            for i in range(len(intervals) - 1):
                start_time = intervals[i]
                end_time = intervals[i + 1]

                # Create a label for the interval
                interval_label = f"{start_time.strftime('%I:%M%p').lstrip('0').lower()}"
                interval_labels.append(interval_label)

                # Filter data for the current interval
                mask = (targets_dt.time >= start_time) & (targets_dt.time < end_time)
                interval_preds = preds[mask]
                interval_targets = targets[mask]

                # Skip empty intervals
                if len(interval_preds) == 0:
                    continue

                # Calculate metrics for this interval
                interval_metrics = self._calculate_metrics(
                    interval_preds, interval_targets, metrics
                )

                # Store results
                for metric_name, metric_value in interval_metrics.items():
                    if metric_name not in results:
                        results[metric_name] = []
                    results[metric_name].append(metric_value)
        else:
            raise ValueError(f"Unsupported period: {period}. Please use 'daily'.")

        # Normalize metrics if requested
        if normalize and len(results) > 0:
            for metric_name, metric_values in results.items():
                min_val = min(metric_values)
                max_val = max(metric_values)
                # Avoid division by zero
                if max_val > min_val:
                    results[metric_name] = [
                        (v - min_val) / (max_val - min_val) for v in metric_values
                    ]
                else:
                    # If all values are the same, set to 0.5 for visualization
                    results[metric_name] = [0.5 for _ in metric_values]

        # Set up the plot
        fig, ax = plt.subplots(figsize=figsize)

        # Secondary axis if requested and we have exactly 2 metrics
        ax2 = None
        if use_secondary_axis and len(results) == 2 and not normalize:
            ax2 = ax.twinx()

        # Plot each metric as a smooth line with markers
        colors = plt.cm.tab10.colors  # Use a colorful palette
        metric_names = list(results.keys())

        for i, (metric_name, metric_values) in enumerate(results.items()):
            color = colors[i % len(colors)]

            # Choose which axis to plot on
            plot_ax = ax
            if i == 1 and ax2 is not None:
                plot_ax = ax2
                plot_ax.set_ylabel(metric_names[1].upper(), color=color)
                plot_ax.tick_params(axis="y", labelcolor=color)

            # Plot with interpolation for smoother lines
            plot_ax.plot(
                range(len(interval_labels)),
                metric_values,
                "-",
                marker="o",
                markersize=6,
                linewidth=2.5,
                label=metric_name.upper(),
                color=color,
                alpha=0.7,
            )  # Semi-transparent to show overlaps

        # Add labels and title
        ax.set_xlabel(x_axis)
        if normalize:
            ax.set_ylabel("Normalized Error (0-1)")
        else:
            ax.set_ylabel(y_axis)

        if title:
            ax.set_title(title)
        else:
            if normalize:
                ax.set_title(
                    f"Normalized Performance Metrics by Time of Day (Interval: {period_interval} min)"
                )
            else:
                ax.set_title(
                    f"Performance Metrics by Time of Day (Interval: {period_interval} min)"
                )

        # Format x-axis for better readability - show labels at wider intervals
        skip_factor = max(
            1, int(len(interval_labels) / 6)
        )  # Show approximately 6 labels on x-axis
        ax.set_xticks(range(0, len(interval_labels), skip_factor))
        ax.set_xticklabels(
            [interval_labels[i] for i in range(0, len(interval_labels), skip_factor)],
            rotation=0,
        )

        # Add legend
        if ax2 is None:
            ax.legend(
                loc="best", framealpha=0.9, fancybox=True, shadow=True, fontsize=11
            )
        else:
            # For dual axis, combine legends from both axes
            lines1, labels1 = ax.get_legend_handles_labels()
            lines2, labels2 = ax2.get_legend_handles_labels()
            ax.legend(
                lines1 + lines2,
                labels1 + labels2,
                loc="best",
                framealpha=0.9,
                fancybox=True,
                shadow=True,
                fontsize=11,
            )

        # Add grid for better readability of values
        ax.grid(True, linestyle="--", alpha=0.6)

        # Style the plot to look more like a financial chart
        ax.spines["top"].set_visible(False)
        if ax2 is None:
            ax.spines["right"].set_visible(False)

        plt.tight_layout()

        # Save if requested
        if save:
            plt.savefig(f"period_visualize.{output_format}", dpi=dpi)

        return fig

    def _calculate_metrics(
        self, preds: np.ndarray, targets: np.ndarray, metrics: List[str]
    ) -> Dict[str, float]:
        """
        Calculate the specified error metric between predictions and targets.

        Args:
            preds (np.ndarray): Model predictions of shape (num_examples, pred_len)
            targets (np.ndarray): Ground truth values of shape (num_examples, pred_len)
            metric (str): Metric to calculate ("mse", "mae", "rmse")

        Returns:
            np.ndarray: Error values with same shape as inputs
        """
        metrics = list(map(str.lower, metrics))

        metric_values = dict()

        if "mse" in metrics:
            metric_values["mse"] = np.mean((preds - targets) ** 2)
        if "mae" in metrics:
            metric_values["mae"] = np.mean(np.abs(preds - targets))
        if "rmse" in metrics:
            metric_values["rmse"] = np.sqrt(np.mean((preds - targets) ** 2))
        if "mape" in metrics:
            epsilon = 1e-10
            metric_values["mape"] = np.mean(
                np.abs((preds - targets) / (targets + epsilon))
            )
        if "smape" in metrics:
            epsilon = 1e-10
            metric_values["smape"] = np.mean(
                2
                * np.abs(preds - targets)
                / (np.abs(preds) + np.abs(targets) + epsilon)
            )
        if "r2" in metrics:
            ssr = ((preds - targets) ** 2).sum()
            sst = ((targets - targets.mean()) ** 2).sum()
            metric_values["r2"] = 1 - ssr / sst

        assert (
            len(metric_values) > 0
        ), f"Invalid metrics: {metrics}. Supported metrics: mse, mae, rmse, mape, r2"

        return metric_values

    def information_coefficient_analysis(
        self, forward_periods: List[int] = [1, 5, 10, 20], rolling_window: int = 20
    ):
        """
        Calculate IC (Information Coefficient) for different forward periods

        Mathematical explanation:
        - Forward Period: Number of time steps ahead that the prediction is targeting
        - IC: Spearman rank correlation between predictions and actual future values
          IC = corr_spearman(prediction_t, actual_{t+forward_period})
        - IC IR (Information Coefficient Information Ratio):
          The ratio of mean IC to standard deviation of IC over time
          IC IR = mean(IC) / std(IC)

        Why it's relevant for forecasting and alpha research:
        - IC measures how well your model ranks outcomes (essential for relative value strategies)
        - IC across different horizons shows decay pattern of your signal
        - IC IR quantifies signal-to-noise ratio - a high IC IR indicates consistent predictive power
        - IC > 0.05 is often considered meaningful in practice for daily forecasts
        - IC analysis helps determine optimal holding periods and trading frequency

        Args:
            forward_periods: List of forward periods to analyze
            rolling_window: Window size for calculating IC IR

        Returns:
            DataFrame with IC metrics by forward period
        """
        ic_metrics = []

        for period in forward_periods:
            # Shift target to align with prediction
            forward_target = self.data["target"].shift(-period)

            # Calculate overall IC - Spearman rank correlation
            ic = stats.spearmanr(
                self.data["prediction"], forward_target, nan_policy="omit"
            )[0]

            # Calculate rolling IC for IR calculation
            self.data[f"rolling_ic_{period}"] = (
                self.data["prediction"]
                .rolling(window=rolling_window)
                .corr(forward_target, method="spearman")
            )

            # Calculate IC IR (Information Ratio of the IC)
            # Mean IC divided by standard deviation of IC
            mean_rolling_ic = self.data[f"rolling_ic_{period}"].mean()
            std_rolling_ic = self.data[f"rolling_ic_{period}"].std()

            # Handle division by zero
            ic_ir = mean_rolling_ic / std_rolling_ic if std_rolling_ic > 0 else np.nan

            # Get time-series stats on the rolling IC
            ic_positive_rate = (self.data[f"rolling_ic_{period}"] > 0).mean()
            ic_significant_rate = (self.data[f"rolling_ic_{period}"] > 0.05).mean()

            ic_metrics.append(
                {
                    "forward_period": period,
                    "ic": ic,
                    "mean_rolling_ic": mean_rolling_ic,
                    "std_rolling_ic": std_rolling_ic,
                    "ic_ir": ic_ir,
                    "ic_positive_rate": ic_positive_rate,
                    "ic_significant_rate": ic_significant_rate,
                }
            )

        return pd.DataFrame(ic_metrics)

    def error_autocorrelation_analysis(self, lags=20, plot=True):
        """
        Analyze autocorrelation in prediction errors

        Args:
            lags: Number of lags to analyze
            plot: Whether to generate visualization

        Returns:
            Series with autocorrelation values by lag
        """
        error_acf = acf(self.data["error"].dropna(), nlags=lags)

        if plot:
            plt.figure(figsize=(12, 6))
            plt.bar(range(len(error_acf)), error_acf)
            plt.axhline(y=0, color="r", linestyle="-", alpha=0.3)
            plt.axhline(
                y=1.96 / np.sqrt(len(self.data)), color="k", linestyle="--", alpha=0.3
            )
            plt.axhline(
                y=-1.96 / np.sqrt(len(self.data)), color="k", linestyle="--", alpha=0.3
            )
            plt.xlabel("Lag")
            plt.ylabel("Autocorrelation")
            plt.title("Error Autocorrelation Function")
            plt.grid(True, alpha=0.3)
            plt.tight_layout()
            plt.show()

        return pd.Series(error_acf, index=range(len(error_acf)))

    def event_study_analysis(self, event_dates, window=5):
        """
        Analyze model performance around specific market events

        Args:
            event_dates: List of event dates
            window: Window size around events to analyze

        Returns:
            DataFrame with performance metrics around events
        """
        event_metrics = []

        for event_date in event_dates:
            # Filter data around event
            event_window = self.data.loc[
                (self.data.index >= event_date - pd.Timedelta(days=window))
                & (self.data.index <= event_date + pd.Timedelta(days=window))
            ]

            # Calculate relative position to event
            event_window = event_window.copy()
            event_window["event_day"] = (
                event_window.index - pd.to_datetime(event_date)
            ).days

            # Group by relative day
            for day, group in event_window.groupby("event_day"):
                if len(group) > 0:
                    mse = mean_squared_error(group["target"], group["prediction"])
                    mae = mean_absolute_error(group["target"], group["prediction"])
                    ic = stats.spearmanr(group["prediction"], group["target"])[0]

                    event_metrics.append(
                        {
                            "event_date": event_date,
                            "event_day": day,
                            "mse": mse,
                            "mae": mae,
                            "ic": ic,
                        }
                    )

        return pd.DataFrame(event_metrics)

    def analyze_forecast_features(self, features=None, n_bins=10, plot=True):
        """
        Analyze model performance conditional on input feature values
        This is especially useful for alpha research to understand what market
        conditions lead to better or worse predictions

        Args:
            features: List of feature columns to analyze, if None use all available columns
            n_bins: Number of quantile bins to divide feature values into
            plot: Whether to generate visualizations

        Returns:
            Dict of DataFrames with performance metrics by feature quantiles
        """
        if features is None:
            # Use all numeric columns except target, prediction, and computed metrics
            exclude_cols = [
                "target",
                "prediction",
                "error",
                "abs_error",
                "squared_error",
                "pct_error",
                "hour",
                "minute",
                "day_of_week",
            ]
            features = [
                col
                for col in self.data.columns
                if col not in exclude_cols
                and np.issubdtype(self.data[col].dtype, np.number)
            ]

        results = {}

        for feature in features:
            if feature not in self.data.columns:
                continue

            # Create quantile bins
            self.data[f"{feature}_bin"] = pd.qcut(
                self.data[feature].fillna(self.data[feature].median()),
                q=n_bins,
                duplicates="drop",
            )

            # Group by feature bins
            feature_metrics = []

            for bin_val, group in self.data.groupby(f"{feature}_bin"):
                if len(group) > 10:  # Ensure enough samples
                    mse = mean_squared_error(group["target"], group["prediction"])
                    mae = mean_absolute_error(group["target"], group["prediction"])
                    ic = stats.spearmanr(group["prediction"], group["target"])[0]
                    dir_acc = np.mean(
                        np.sign(group["prediction"]) == np.sign(group["target"])
                    )

                    feature_metrics.append(
                        {
                            "feature": feature,
                            "bin": bin_val,
                            "bin_center": bin_val.mid,
                            "mse": mse,
                            "mae": mae,
                            "ic": ic,
                            "dir_acc": dir_acc,
                            "samples": len(group),
                        }
                    )

            # Convert to DataFrame
            feature_df = pd.DataFrame(feature_metrics).sort_values("bin_center")
            results[feature] = feature_df

            # Generate plot if requested
            if plot and not feature_df.empty:
                fig, ax1 = plt.subplots(figsize=(10, 6))

                # Plot MSE on left axis
                ax1.plot(
                    range(len(feature_df)),
                    feature_df["mse"],
                    "o-",
                    color="blue",
                    label="MSE",
                )
                ax1.set_xlabel(feature)
                ax1.set_ylabel("MSE", color="blue")
                ax1.tick_params(axis="y", labelcolor="blue")

                # Plot IC on right axis
                ax2 = ax1.twinx()
                ax2.plot(
                    range(len(feature_df)),
                    feature_df["ic"],
                    "o-",
                    color="red",
                    label="IC",
                )
                ax2.set_ylabel("IC", color="red")
                ax2.tick_params(axis="y", labelcolor="red")

                # Set x-ticks
                bin_labels = [f"{b.left:.2f}-{b.right:.2f}" for b in feature_df["bin"]]
                plt.xticks(range(len(feature_df)), bin_labels, rotation=45)

                plt.title(f"Model Performance by {feature} Quantiles")
                plt.tight_layout()
                plt.show()

        return results

    def get_torch_preds(
        self,
        model: nn.Module,
        dataset: Any,  # ForecastingDataset
        batch_size: int = 128,
        channel: int = 0,
        device: str = "cuda" if torch.cuda.is_available() else "cpu",
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Generate predictions from a PyTorch model and ForecastingDataset.
        Args:
            model (nn.Module): PyTorch model
            dataset (ForecastingDataset): ForecastingDataset object
            batch_size (int): Batch size for DataLoader
            device (str): Device to run model on ('cuda' or 'cpu')
        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]:
                (predictions, targets, target_datetimes)
        """
        all_preds = []
        all_targets = []
        all_targets_dt = []
        loader = DataLoader(
            dataset, batch_size=batch_size, shuffle=False, drop_last=False
        )

        # Set model to evaluation mode
        model.eval()

        with torch.no_grad():
            for batch in loader:
                x, y, x_dt, y_dt = (
                    batch  # x: (batch_size, num_feats, seq_len), y: (batch_size, num_feats, pred_len)
                )

                # Add batch dimension if needed
                if x.dim() == 2:
                    x = x.unsqueeze(0)
                    y = y.unsqueeze(0)
                    x_dt = x_dt.unsqueeze(0)
                    y_dt = y_dt.unsqueeze(0)

                # Compute predictions
                x = x.to(device)
                preds = model(x)  # preds: (batch_size, num_feats, pred_len)

                # Append to lists
                all_preds.append(preds)
                all_targets.append(y)
                all_targets_dt.append(y_dt)

        # Stack all predictions, targets, and datetimes
        preds = torch.cat(all_preds).detach().cpu().numpy()
        targets = torch.cat(all_targets).detach().cpu().numpy()
        targets_dt = tensor_to_datetime(
            timestamp_tensor=torch.cat(all_targets_dt).detach().cpu(), batch_mode=True
        )

        # Select the target channel
        preds = preds[:, channel, :]
        targets = targets[:, channel, :]

        return preds, targets, targets_dt


if __name__ == "__main__":
    # Step 1: Find and initialize the optimal contract
    from optrade.data.contracts import Contract

    contract = Contract.find_optimal(
        root="AAPL",
        right="C",  # Call option
        start_date="20230103",  # First trading day of 2023
        target_tte=30,  # Desired expiration: 30 days
        tte_tolerance=(20, 40),  # Min 20, max 40 days expiration
        interval_min=1,  # Data requested at 1-min level
        moneyness="ATM",  # At-the-money option
        dev_mode=True,
    )

    # Step 2: Load market data (NBBO quotes and OHLCV)
    df = contract.load_data(dev_mode=True)

    # Step 3: Transform raw data into ML-ready features
    from optrade.data.features import transform_features

    core_feats = [
        "option_returns",  # Option price returns
        "stock_returns",  # Underlying stock returns
        "moneyness",  # Log(S/K)
        "option_lob_imbalance",  # Order book imbalance
        "stock_quote_spread",  # Bid-ask spread normalized
    ]
    tte_feats = ["sqrt", "exp_decay"]  # Time-to-expiration features
    datetime_feats = ["minute_of_day", "hour_of_week"]
    vol_feats = ["rolling_volatility", "vol_ratio"]
    rolling_volatility_range = [20, 60] # Rolling 20min, and 60min past volatility values

    data = transform_features(
        df=df,
        core_feats=core_feats,
        tte_feats=tte_feats,  # Time-to-expiration features
        datetime_feats=datetime_feats,  # Time features
        vol_feats=vol_feats,
        rolling_volatility_range=rolling_volatility_range,
        root = contract.root,
        right=contract.right,
        strike=contract.strike,
        exp=contract.exp,
        keep_datetime=True,
    )

    # Step 4: Create dataset for time series forecasting
    from optrade.data.forecasting import ForecastingDataset
    from torch.utils.data import DataLoader

    target_channels = ["option_returns"]
    torch_dataset = ForecastingDataset(
        data=data,
        seq_len=100,  # 100-minute lookback window
        pred_len=10,  # 10-minute forecast horizon
        target_channels=target_channels,
    )

    # Test the _get_torch_predictions method
    import torch.nn as nn

    model = nn.Linear(100, 10)

    # Randomly initialize weights
    with torch.no_grad():
        model.weight = nn.Parameter(torch.randn_like(model.weight))
        model.bias = nn.Parameter(torch.randn_like(model.bias))

    visualizer = Analyzer()

    # Test period_visualize
    fig = visualizer.period_visualize(
        period="daily",
        period_interval=5,
        model=model,
        dataset=torch_dataset,
        metrics=["mse", "mae", "r2", "mape"],
        batch_size=128,
        x_axis="Time (min)",
        y_axis="MSE & MAE",
        title="Model Performance by Time of Day",
        figsize=(12, 6),
        dpi=300,
        save=False,
        normalize=True,
    )
    plt.show()



import torch
import torch.nn as nn
from torch import Tensor

from optrade.models.utils.utils import Norm
from optrade.models.utils.utils import Reshape
from optrade.models.utils.revin import RevIN
from optrade.models.utils.patcher import Patcher
from optrade.models.utils.pos_enc import PositionalEncoding
from optrade.models.utils.weight_init import xavier_init

from typing import Optional, List


class Model(nn.Module):
    def __init__(
        self,
        num_enc_layers: int,
        d_model: int,
        d_ff: int,
        num_heads: int,
        input_channels: List[str],
        seq_len: int,
        pred_len: int,
        attn_dropout: float = 0.0,
        ff_dropout: float = 0.0,
        pred_dropout: float = 0.0,
        batch_first: bool = True,
        norm_mode: str = "batch1d",
        revin: bool = True,
        revout: bool = True,
        revin_affine: bool = True,
        eps_revin: float = 1e-5,
        patch_dim: int = 16,
        stride: int = 1,
        return_head: bool = True,
        head_type: str = "linear",
        channel_independent: bool = False,  # Head only
        target_channels: Optional[list] = None,  # Head only
    ) -> None:
        super(Model, self).__init__()

        # Parameters
        self.num_patches = int((seq_len - patch_dim) / stride) + 2
        self.num_channels = len(input_channels)
        self.input_channels = input_channels
        self.eps_revin = eps_revin
        self.revin_affine = revin_affine
        self.target_channels = target_channels
        self.revout = revout

        # Initialize layers
        if revin:
            self._init_revin()
        else:
            self._revin = None
            self.revout = None

        self.patcher = Patcher(patch_dim, stride)
        self.pos_enc = PositionalEncoding(patch_dim, d_model, self.num_patches)
        self.backbone = PatchTSTBackbone(
            input_channels=input_channels,
            num_enc_layers=num_enc_layers,
            d_model=d_model,
            d_ff=d_ff,
            num_heads=num_heads,
            num_channels=self.num_channels,
            num_patches=self.num_patches,
            pred_len=pred_len,
            attn_dropout=attn_dropout,
            ff_dropout=ff_dropout,
            pred_dropout=pred_dropout,
            batch_first=batch_first,
            norm_mode=norm_mode,
            return_head=return_head,
            head_type=head_type,
            channel_independent=channel_independent,
            target_channels=target_channels,
        )

        # Weight initialization
        self.apply(xavier_init)

    def _init_revin(self):
        self._revin = True
        self.revin = RevIN(
            num_channels=self.num_channels,
            input_channels=self.input_channels,
            eps=self.eps_revin,
            affine=self.revin_affine,
            target_channels=self.target_channels,
        )

    def forward(self, x, y=None, ch_ids=None):

        # RevIN
        if self._revin:
            x = self.revin(x, mode="norm")

        # Patcher
        x = self.patcher(x)

        # Project + Positional Encoding
        x = self.pos_enc(x)

        # Transformer + Linear Head
        x = self.backbone(x, y, ch_ids)

        # RevOUT
        if self.revout:
            x = self.revin(x, mode="denorm")

        return x


class PatchTSTBackbone(nn.Module):
    def __init__(
        self,
        num_enc_layers,
        d_model,
        d_ff,
        num_heads,
        num_channels,
        num_patches,
        pred_len,
        input_channels,
        attn_dropout=0.0,
        ff_dropout=0.0,
        pred_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
        return_head=True,
        head_type="linear",
        channel_independent=False,
        target_channels: Optional[list] = None,
    ):
        super(PatchTSTBackbone, self).__init__()

        # Parameters
        self.num_channels = num_channels
        self.num_patches = num_patches
        self.d_model = d_model
        self.return_head = return_head
        self.pred_len = pred_len
        self.target_channels = target_channels
        if target_channels is not None:
            self.target_channels_idx = [
                input_channels.index(ch) for ch in target_channels
            ]

        # Encoder
        self.enc = nn.Sequential(
            *(
                EncoderBlock(
                    d_model,
                    d_ff,
                    num_heads,
                    num_channels,
                    num_patches,
                    attn_dropout,
                    ff_dropout,
                    batch_first,
                    norm_mode,
                )
                for i in range(num_enc_layers)
            )
        )

        # Prediction head
        self.flatten = nn.Flatten(start_dim=-2)
        num_output_channels = (
            len(target_channels) if target_channels is not None else num_channels
        )
        if head_type == "linear":
            if channel_independent:
                self.head = nn.Linear(num_patches * d_model, pred_len)
            elif target_channels is not None or not channel_independent:
                self.head = nn.Sequential(
                    Reshape(-1, num_output_channels * num_patches * d_model),
                    nn.Linear(
                        num_output_channels * num_patches * d_model,
                        num_output_channels * pred_len,
                    ),
                    Reshape(-1, num_output_channels, pred_len),
                )
        elif head_type == "mlp":
            self.head = nn.Sequential(
                nn.Linear(num_patches * d_model, num_patches * d_model),
                nn.GELU(),
                nn.Dropout(pred_dropout),
                nn.Linear(num_patches * d_model, pred_len),
            )
        else:
            raise ValueError(f"Invalid head type: {head_type}")

    def forward(
        self, x: torch.Tensor, y: torch.Tensor = None, ch_ids: torch.Tensor = None
    ) -> torch.Tensor:

        # Encoding
        batch_size = x.shape[0]
        x = x.view(
            batch_size * self.num_channels, self.num_patches, -1
        )  # (batch_size * num_channels, num_patches, d_model)
        x = self.enc(x)  # (batch_size * num_channels, num_patches, d_model)

        if self.return_head:
            x = x.reshape(
                batch_size, self.num_channels, self.num_patches * self.d_model
            )

            if self.target_channels_idx is not None:
                x = x[:, self.target_channels_idx, :]

            out = self.head(x)
        else:
            out = x.view(batch_size, self.num_channels, self.num_patches, -1)

        return out


class SupervisedHead(nn.Module):
    def __init__(self, linear_dim, pred_len, dropout=0.0):
        super().__init__()
        """
        Flattens and applies a linear layer to each channel independently to form a prediction.
        Args:
            num_channels (int): The number of channels in the input.
            linear_dim (int): The dimension of the linear layer, should be num_patches * d_model.
            pred_len (int): The length of the forecast window.
            dropout (float): The dropout value.
        """

        self.dropout = nn.Dropout(dropout)
        self.linear = nn.Linear(linear_dim, pred_len)

    def forward(self, x) -> torch.Tensor:
        """
        Applies a linear layer to each channel independently to form a prediction, optional dropout.
        Args:
            x (torch.Tensor): The input of shape (batch_size, num_channels, num_patches, d_model)
        Returns:
            x (torch.Tensor): The output of shape (batch_size, num_channels, pred_len).
        """
        x = self.linear(x)
        x = self.dropout(x)
        return x


class EncoderBlock(nn.Module):
    """
    Args:
        d_model: The embedding dimension.
        num_heads: The number of heads in the multi-head attention models.
        dropout: The dropout value.
        batch_first: If True, then the input and output tensors are provided as (batch, seq, feature).
        norm: The type of normalization to use. Either "batch1d", "batch2d", or "layer".
    """

    def __init__(
        self,
        d_model,
        d_ff,
        num_heads,
        num_channels,
        num_patches,
        attn_dropout=0.0,
        ff_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
    ):
        super(EncoderBlock, self).__init__()

        # Layers
        self.attn = _MultiheadAttention(
            num_heads=num_heads,
            d_model=d_model,
            dropout=attn_dropout,
            batch_first=batch_first,
        )
        self.ff = nn.Sequential(
            nn.Linear(d_model, d_ff),
            nn.GELU(),
            nn.Dropout(ff_dropout),
            nn.Linear(d_ff, d_model),
        )

        # Normalization
        self.norm = Norm(norm_mode, num_channels, num_patches, d_model)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: Input tensor of shape (batch_size, num_patches, d_model).
        Returns:
            fc_out: Output of the transformer block, a tensor of shape (batch_size, num_patches, d_model).
        """

        # Multihead Attention -> Add & Norm
        attn_out, _ = self.attn(x, x, x)
        attn_norm = self.norm(
            attn_out + x
        )  # Treat the input as the query, key and value for MHA.

        # Feedforward layer -> Add & Norm
        fc_out = self.ff(attn_norm)
        fc_norm = self.norm(fc_out + attn_out)

        return fc_norm


class _MultiheadAttention(nn.Module):
    """
    Multihead Attention mechanism from the Vanilla Transformer, with some preset parameters for the PatchTST model.
    """

    def __init__(self, num_heads: int, d_model: int, dropout=0.0, batch_first=True):
        super(_MultiheadAttention, self).__init__()

        # Layers
        self.attn = nn.MultiheadAttention(
            embed_dim=d_model,
            num_heads=num_heads,
            dropout=dropout,
            batch_first=batch_first,
        )

    def forward(
        self, Q: torch.Tensor, K: torch.Tensor, V: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            Q: Query embedding of shape: (batch_size, num_patches, d_model).
            K: Key embedding of shape (batch_size, num_patches, d_model).
            V: Value embedding of shape (batch_size, num_patches, d_model).
            batch_size: The batch size.
            num_patches: The sequence length.
            d_model: The embedding dimension.
        Returns:
            x: The output of the attention layer of shape (batch_size, num_patches, d_model).
        """
        return self.attn(query=Q, key=K, value=V, need_weights=False)


if __name__ == "__main__":
    batch_size = 64
    num_channels = 7
    seq_len = 512
    patch_dim = 16
    patch_stride = 8
    pred_len = 96

    model = Model(
        num_enc_layers=3,
        d_model=128,
        d_ff=512,
        num_heads=4,
        num_channels=num_channels,
        seq_len=seq_len,
        pred_len=pred_len,
        attn_dropout=0.0,
        ff_dropout=0.0,
        pred_dropout=0.0,
        batch_first=True,
        norm_mode="batch1d",
        revin=True,
        revout=False,
        revin_affine=True,
        eps_revin=1e-5,
        patch_dim=patch_dim,
        stride=patch_stride,
        return_head=True,
        head_type="linear",
        channel_independent=False,
        target_channels=[1, 3, 0],
    )

    x = torch.randn(batch_size, num_channels, seq_len)  # (B, M, L)
    out = model(x)
    print(f"Input shape: {x.shape}, Output shape: {out.shape}")



import warnings
import copy
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from pydantic import BaseModel

warnings.filterwarnings("ignore", message="h5py not installed")

from typing import List

# Supervised Models
from optrade.models.pytorch.patchtst import Model as PatchTST
from optrade.models.pytorch.recurrent import Model as RecurrentModel
from optrade.models.pytorch.linear import Model as Linear
from optrade.models.pytorch.dlinear import Model as DLinear
from optrade.models.pytorch.tsmixer import Model as TSMixer
from optrade.models.pytorch.emforecaster import Model as EMForecaster

# Optimizers and Schedulers
from torch import optim

from optrade.dev.utils.schedulers import WarmupCosineSchedule
from torch.optim.lr_scheduler import CosineAnnealingLR, OneCycleLR


def get_model(
    args: BaseModel,
    input_channels: List[str],
    target_channels_idx: List[int],
) -> nn.Module:
    if args.exp.model_id == "PatchTST":
        model = PatchTST(
            input_channels=input_channels,
            num_enc_layers=args.patchtst.num_enc_layers,
            d_model=args.patchtst.d_model,
            d_ff=args.patchtst.d_ff,
            num_heads=args.patchtst.num_heads,
            seq_len=args.data.seq_len,
            pred_len=args.data.pred_len,
            attn_dropout=args.patchtst.attn_dropout,
            ff_dropout=args.patchtst.ff_dropout,
            pred_dropout=args.patchtst.pred_dropout,
            batch_first=args.train.batch_first,
            norm_mode=args.train.norm_mode,
            revin=args.train.revin,
            revout=args.train.revout,
            revin_affine=args.train.revin_affine,
            eps_revin=args.train.eps_revin,
            patch_dim=args.patchtst.patch_dim,
            stride=args.patchtst.patch_stride,
            return_head=args.train.return_head,
            head_type=args.train.head_type,
            channel_independent=args.train.channel_independent,  # Head only
            target_channels=target_channels_idx,  # Head only
        )
    elif args.exp.model_id == "RecurrentModel":
        model = RecurrentModel(
            d_model=args.rnn.d_model,
            num_enc_layers=args.rnn.num_enc_layers,
            pred_len=args.data.pred_len,
            backbone_id=args.rnn.backbone_id,
            bidirectional=args.rnn.bidirectional,
            dropout=args.train.dropout,
            seq_len=args.data.seq_len,
            patching=args.data.patching,
            patch_dim=args.rnn.patch_dim,
            patch_stride=args.rnn.patch_stride,
            num_channels=args.data.num_channels,
            head_type=args.train.head_type,
            norm_mode=args.train.norm_mode,
            revin=args.train.revin,
            revout=args.train.revout,
            revin_affine=args.train.revin_affine,
            eps_revin=args.train.eps_revin,
            last_state=args.rnn.last_state,
            avg_state=args.rnn.avg_state,
            return_head=args.train.return_head,
            channel_independent=args.train.channel_independent,
            target_channels=target_channels_idx,
        )
    elif args.exp.model_id == "Linear":
        model = Linear(
            seq_len=args.data.seq_len,
            pred_len=args.data.pred_len,
            num_channels=args.data.num_channels,
            norm_mode=args.train.norm_mode,
            revin=args.train.revin,
            revout=args.train.revout,
            revin_affine=args.train.revin_affine,
            eps_revin=args.train.eps_revin,
            channel_independent=args.train.channel_independent,
            target_channels=target_channels_idx,
        )
    elif args.exp.model_id == "DLinear":
        model = DLinear(
            task=args.exp.task,
            seq_len=args.data.seq_len,
            pred_len=args.data.pred_len,
            num_channels=args.data.num_channels,
            num_classes=args.data.num_classes,
            moving_avg=args.dlinear.moving_avg,
            individual=args.dlinear.individual,
            return_head=args.train.return_head,
            revin=args.train.revin,
            revout=args.train.revout,
            revin_affine=args.train.revin_affine,
            eps_revin=args.train.eps_revin,
            target_channels=target_channels_idx,
        )
    elif args.exp.model_id == "EMForecaster":
        model = EMForecaster(
            args=args,
            seed=args.exp.seed,
            seq_len=args.data.seq_len,
            pred_len=args.data.pred_len,
            num_channels=args.data.num_channels,
            revin=args.train.revin,
            revout=args.train.revout,
            revin_affine=args.train.revin_affine,
            eps_revin=args.train.eps_revin,
            patch_model_id=args.emf.patch_model_id,
            patch_norm=args.emf.patch_norm,
            patch_act=args.emf.patch_act,
            patch_dim=args.emf.patch_dim,
            patch_stride=args.emf.patch_stride,
            patch_embed_dim=args.emf.patch_embed_dim,
            pos_enc=args.emf.pos_enc,
            return_head=args.train.return_head,
            target_channels=target_channels_idx,
        )
    elif args.exp.model_id == "TSMixer":
        model = TSMixer(
            seq_len=args.data.seq_len,
            pred_len=args.data.pred_len,
            num_enc_layers=args.tsmixer.num_enc_layers,
            d_model=args.tsmixer.d_model,
            num_channels=args.data.num_channels,
            dropout=args.train.dropout,
            revin=args.train.revin,
            revin_affine=args.train.revin_affine,
            revout=args.train.revout,
            eps_revin=args.train.eps_revin,
            return_head=args.train.return_head,
            target_channels=target_channels_idx,
            channel_independent=args.train.channel_independent,
        )
    else:
        raise ValueError("Please select a valid model_id.")
    return model


def get_optim(args, model):
    optimizer_classes = {"adam": optim.Adam, "adamw": optim.AdamW}

    if args.train.optimizer not in optimizer_classes:
        raise ValueError("Please select a valid optimizer.")
    optimizer_class = optimizer_classes[args.train.optimizer]

    param_groups = exclude_weight_decay(
        model, args
    )  # Exclude bias and normalization parameters from weight decay

    optimizer = optimizer_class(param_groups)  # Set optimizer

    return optimizer


def exclude_weight_decay(model, args):
    # Separate parameters into those that will use weight decay and those that won't
    decay_params = []
    no_decay_params = []

    for name, param in model.named_parameters():
        if param.requires_grad:
            if "bias" in name or isinstance(
                param, (nn.BatchNorm1d, nn.BatchNorm2d, nn.LayerNorm)
            ):
                no_decay_params.append(param)
            else:
                decay_params.append(param)

    # Create parameter groups
    param_groups = [
        {"params": decay_params, "weight_decay": args.train.weight_decay},
        {"params": no_decay_params, "weight_decay": 0.0},
    ]

    return param_groups


def get_scheduler(args: BaseModel, optimizer, num_batches: int = 0):
    if args.train.scheduler == "cosine":
        scheduler = CosineAnnealingLR(
            optimizer,
            T_max=args.train.epochs,
            eta_min=args.train.lr * 1e-2,
            last_epoch=args.scheduler.last_epoch,
        )
    elif args.train.scheduler == "onecycle":
        scheduler = OneCycleLR(
            optimizer=optimizer,
            steps_per_epoch=num_batches,
            pct_start=args.scheduler.pct_start,
            epochs=args.train.epochs,
            max_lr=args.train.lr,
        )
    else:
        raise ValueError("Please select a valid scheduler_type.")
    return scheduler


def get_criterion(args: BaseModel) -> nn.Module:
    if args.train.criterion == "MSE":
        return nn.MSELoss()
    elif args.train.criterion == "SmoothL1":
        return nn.SmoothL1Loss()
    elif args.train.criterion == "BCE":
        return nn.BCEWithLogitsLoss()
    elif args.train.criterion == "BCE_normal":
        return nn.BCELoss()
    elif args.train.criterion == "CE":
        return nn.CrossEntropyLoss()
    else:
        raise ValueError("Please select a valid criterion.")


def forward_pass(args, model, batch, model_id, device):
    if model_id in {
        "PatchTST",
        "TSMixer",
        "RecurrentModel",
        "Linear",
        "DLinear",
        "ModernTCN",
        "TimesNet",
        "EMForecaster",
    }:
        x = batch[0]
        x = x.to(device)

        if args.data.difference_input:
            x = x.diff(dim=1)
            # Pad the last dimension with the last value of the sequence
            x = torch.cat([x, x[:, -1].unsqueeze(1)], dim=1)

        output = model(x)
    else:
        raise ValueError("Please select a valid model_id.")

    return output


def compute_loss(output, batch, criterion, model_id, args, device):
    if model_id in {
        "PatchTST",
        "TSMixer",
        "RecurrentModel",
        "Linear",
        "DLinear",
        "ModernTCN",
        "TimesNet",
        "EMForecaster",
    }:
        target = batch[1].to(device)
        loss = criterion(output, target)
    else:
        raise ValueError("Please select a valid model_id.")

    return loss


def model_update(model, loss, optimizer, model_id):
    if model_id in {
        "PatchTST",
        "TSMixer",
        "RecurrentModel",
        "Linear",
        "DLinear",
        "ModernTCN",
        "TimesNet",
        "EMForecaster",
    }:
        loss.backward()
        optimizer.step()
    else:
        raise ValueError(f"{model_id} is not a valid model_id.")


def check_gradients(model, threshold_low=1e-5, threshold_high=1e2):
    vanishing = []
    exploding = []
    normal = []

    for name, param in model.named_parameters():
        if param.grad is not None:
            grad_norm = param.grad.norm().item()
            if grad_norm < threshold_low:
                vanishing.append((name, grad_norm))
            elif grad_norm > threshold_high:
                exploding.append((name, grad_norm))
            else:
                normal.append((name, grad_norm))

    print(f"Gradient statistics:")
    print(
        f"  Total parameters with gradients: {len(vanishing) + len(exploding) + len(normal)}"
    )
    print(
        f"  Vanishing gradients: {len(vanishing)} ({len(vanishing) / (len(vanishing) + len(exploding) + len(normal)) * 100:.2f}%)"
    )
    print(
        f"  Exploding gradients: {len(exploding)} ({len(exploding) / (len(vanishing) + len(exploding) + len(normal)) * 100:.2f}%)"
    )
    print(
        f"  Normal gradients: {len(normal)} ({len(normal) / (len(vanishing) + len(exploding) + len(normal)) * 100:.2f}%)"
    )

    if vanishing:
        print("\nVanishing gradients:")
        for name, grad_norm in vanishing[:10]:  # Print first 10
            print(f"  {name}: {grad_norm}")
        if len(vanishing) > 10:
            print(f"  ... and {len(vanishing) - 10} more")

    if exploding:
        print("\nExploding gradients:")
        for name, grad_norm in exploding[:10]:  # Print first 10
            print(f"  {name}: {grad_norm}")
        if len(exploding) > 10:
            print(f"  ... and {len(exploding) - 10} more")

    # Compute and print gradient statistics
    all_grads = [
        param.grad.norm().item()
        for name, param in model.named_parameters()
        if param.grad is not None
    ]
    if all_grads:
        print("\nGradient norm statistics:")
        print(f"  Mean: {np.mean(all_grads):.6f}")
        print(f"  Median: {np.median(all_grads):.6f}")
        print(f"  Std: {np.std(all_grads):.6f}")
        print(f"  Min: {np.min(all_grads):.6f}")
        print(f"  Max: {np.max(all_grads):.6f}")



from typing import Optional, Dict, Any

from sklearn.base import BaseEstimator
from sklearn.ensemble import (
    RandomForestRegressor, GradientBoostingRegressor,
    RandomForestClassifier, GradientBoostingClassifier
)
from sklearn.linear_model import Lasso, Ridge, RidgeClassifier, LogisticRegression
from sklearn.multioutput import MultiOutputRegressor

from xgboost import XGBRegressor, XGBClassifier
from catboost import CatBoostRegressor, CatBoostClassifier

def get_sklearn_model(
    model_id: str,
    target_type: str,
    model_args: Optional[Dict[str, Any]] = None
) -> BaseEstimator:
    """
    Instantiate and return a scikit-learn (or compatible) estimator by ID,
    unpacking all kwargs from a dict.

    Args:
        model_id (str): One of
            "random_forest", "gradient_boosting", "lasso", "linear",
            "xgboost", or "catboost".
        target_type (str):
            - "multistep": forecasting a vector of future values
            - "average": forecasting a single continuous value
            - "average_direction": forecasting a binary label (0 or 1)
        model_args (Optional[Dict[str,Any]]):
            All keyword args to pass to the underlying constructor, e.g.
            {"n_estimators":200, "max_depth":5}.
            If None, defaults to {}.

    Returns:
        BaseEstimator: The configured estimator.  If target_type=="multistep",
        wraps in MultiOutputRegressor.
    """
    args = {} if model_args is None else model_args.copy()
    args.setdefault("random_state", 1995)

    if model_id == "rf":
        Est = (
            RandomForestClassifier
            if target_type == "average_direction"
            else RandomForestRegressor
        )
        base = Est(**args)

    elif model_id == "gb":
        Est = (
            GradientBoostingClassifier
            if target_type == "average_direction"
            else GradientBoostingRegressor
        )
        base = Est(**args)

    elif model_id == "lasso":
        if target_type == "average_direction":
            base = LogisticRegression
        else:
            base = Lasso(**args)

    elif model_id == "ridge":
        if target_type == "average_direction":
            base = RidgeClassifier(**args)
        else:
            base = Ridge(**args)

    elif model_id == "xgb":
        Est = XGBClassifier if target_type == "average_direction" else XGBRegressor
        base = Est(**args)

    elif model_id == "cb":
        # CatBoost expects "iterations" not "n_estimators"
        cb_args = args.copy()
        if "n_estimators" in cb_args:
            cb_args["iterations"] = cb_args.pop("n_estimators")
        # rename random_state → random_seed
        cb_args.setdefault("random_seed", cb_args.pop("random_state"))
        Est = (
            CatBoostClassifier
            if target_type == "average_direction"
            else CatBoostRegressor
        )
        # silence the verbose by default
        cb_args.setdefault("verbose", False)
        base = Est(**cb_args)

    else:
        raise ValueError(f"Unknown model_id: {model_id}")

    if target_type == "multistep" and model_id in ["rf", "gb", "xgb", "cb"]:
        return MultiOutputRegressor(base)

    return base



import json
import warnings
import random
import time
import neptune
import joblib
from typing import Optional, List, Tuple, Union
from pathlib import Path
from rich.console import Console
from rich.table import Table
from sklearn.base import BaseEstimator
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.model_selection import TimeSeriesSplit

# Torch
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import torch.optim as optim
from torch.optim.lr_scheduler import _LRScheduler

# Custom Modules
from optrade.data.contracts import get_contract_datasets
from optrade.data.forecasting import get_forecasting_dataset, get_forecasting_loaders
from optrade.utils.train import EarlyStopping
from optrade.utils.misc import format_time_dynamic, generate_random_id
from optrade.exp.evaluate import get_metrics
from optrade.data.universe import Universe

warnings.filterwarnings(
    "ignore", message="h5py not installed, hdf5 features will not be supported."
)


class Experiment:
    def __init__(
        self,
        log_dir: Union[str, Path] = "logs",
        logging: str = "offline",
        seed: int = 42,
        ablation_id: Optional[int] = None,
        exp_id: str = generate_random_id(length=6),
        neptune_project_name: Optional[str] = None,
        neptune_api_token: Optional[str] = None,
        download_only: bool = False,
    ) -> None:
        """

        Experiment class for training and evaluating forecasting models in PyTorch.

        Args:
            logdir: The directory to save logs.
            logging: The logging method to use. Options: {"offline", "neptune"}.
            seed: The random seed to use for reproducibility.
            ablation_id: The ablation ID.
            exp_id: The experiment ID.
            neptune_project_name: The Neptune project name.
            neptune_api_token: The Neptune API token.

        Returns:
            None
        """
        self.start_time = time.time()

        # Rich Console
        self.ctx = Console()
        self.seed = seed
        self.logging = logging
        self.neptune_project_name = neptune_project_name
        self.neptune_api_token = neptune_api_token
        self.exp_id = exp_id
        self.log_dir = Path(log_dir)

        if ablation_id:
            self.log_dir = self.log_dir / "logs" / exp_id / str(ablation_id)
        else:
            self.log_dir = self.log_dir / "logs" / exp_id

        # Set up logging directory
        if not download_only:
            self.init_logger(exp_id=exp_id)

        self.set_seed()

    def save_logs(self):
        # Stop Logger
        if self.logging == "neptune":
            self.logger.stop()
        elif self.logging == "offline":
            # Save offline logging to JSON file
            with open(self.log_file, "w") as f:
                json.dump(self.logger, f, indent=2)
        else:
            raise ValueError(f"Invalid logging method: {self.logging}.")

    # Initialize device
    def init_device(self, gpu_id: int = 0, mps: bool = False) -> None:
        """
        Initialize CUDA (or MPS) devices.

        Args:
            gpu_id: The GPU ID to use.
            mps: Whether to use MPS (Metal Performance Shaders) for macOS.

        Returns:
            None
        """
        if mps:
            self.device = torch.device(
                "mps" if torch.backends.mps.is_available() else "cpu"
            )
            self.print_master("MPS hardware acceleration activated.")
        elif torch.cuda.is_available():
            self.device = torch.device(f"cuda:{gpu_id}")
        else:
            self.device = torch.device("cpu")
            self.print_master("CUDA not available. Running on CPU.")

    def init_logger(
        self,
        exp_id: str,
    ) -> None:
        """
        Initialize the logger.

        Args:
            logdir: The directory to save logs.
            ablation_id: The ablation ID.
            exp_id: The experiment ID.

        Returns:
            None
        """

        if self.logging == "neptune":
            # Initialize Neptune run with the time-based ID
            self.logger = neptune.init_run(
                project=self.neptune_project_name,
                api_token=self.neptune_api_token,
                custom_run_id=exp_id,
            )
            self.print_master("Neptune logger initialized.")
        elif self.logging == "offline":
            self.logger = dict()
            self.log_file = self.log_dir / "log.json"

            if Path(self.log_dir).exists():
                self.print_master(f"Using existing log directory: {self.log_dir}")
            else:
                Path(self.log_dir).mkdir(parents=True, exist_ok=True)
                self.print_master(f"Created new log directory: {self.log_dir}")

            self.print_master("Offline logger initialized.")
        else:
            raise ValueError(f"Invalid logging method: {self.logging}.")

    def init_loaders(
        self,
        root: str,
        start_date: str,
        end_date: str,
        contract_stride: int,
        interval_min: int,
        right: str,
        target_tte: int,
        tte_tolerance: Tuple[int, int],
        moneyness: str,
        train_split: float,
        val_split: float,
        seq_len: int,
        pred_len: int,
        scaling: bool = False,
        dtype: str = "float32",
        core_feats: List[str] = ["option_returns"],
        tte_feats: Optional[List[str]] = None,
        datetime_feats: Optional[List[str]] = None,
        vol_feats: Optional[List[str]] = None,
        rolling_volatility_range: Optional[List[int]] = None,
        keep_datetime: bool = False,
        target_channels: Optional[List[str]] = None,
        target_type: str = "multistep",
        strike_band: Optional[float] = 0.05,
        volatility_type: Optional[str] = "period",
        volatility_scaled: bool = False,
        volatility_scalar: Optional[float] = 1.0,
        batch_size: int = 32,
        shuffle: bool = True,
        drop_last: bool = False,
        num_workers: int = 4,
        prefetch_factor: Optional[int] = 2,
        pin_memory: bool = torch.cuda.is_available(),
        persistent_workers: bool = True,
        clean_up: bool = False,
        offline: bool = False,
        save_dir: Optional[str] = None,
        verbose: bool = False,
        validate_contracts: bool = True,
        modify_contracts: bool = False,
        dev_mode: bool = False,
        download_only: bool = False,
    ) -> None:
        """

        Initializes the data loaders for training, validation, and testing.

        Args:
            root: The root directory containing the data.
            start_date: The start date for the data in YYYYMMDD format.
            end_date: The end date for the data in YYYYMMDD format.
            contract_stride: The stride for the contracts.
            interval_min: The interval in minutes for the data.
            right: The option type (C for call, P for put).
            target_tte: The target time to expiration in minutes.
            moneyness: The moneyness type (ATM, OTM, ITM).
            train_split: The fraction of data to use for training.
            val_split: The fraction of data to use for validation.
            strike_band: The band to select the strike price for OTM or ITM options.
            volatility_type: The type of historical volatility to use.
            volatility_scaled: Whether to scale strike selection by the volatility.
            volatility_scalar: The scalar to multiply the volatility.
            validate_contracts: Whether to validate contracts by requesting the data from ThetaData API.
            modify_contracts: Whether to overwite contracts .pkl files if certain contracts are invalid.
            seq_len: The sequence length of the lookback widow.
            pred_len: The prediction length of the forecast window.
            scaling: Whether to apply normalization.
            core_feats: The core features to use for the model.
            tte_feats: The time-to-expiration features to use for the model.
            datetime_feats: The datetime features to use for the model.
            keep_datetime: Whether to keep the datetime column in the dataset.
            target_channels: The target channels to use for the model.
            target_type: The type of target. Options: "multistep", "average", or "average_direction".
            batch_size: The batch size for the data loaders.
            shuffle: Whether to shuffle the data.
            drop_last: Whether to drop the last batch if it is smaller than the batch size.
            num_workers: The number of workers for the data loaders.
            prefetch_factor: The number of batches to prefetch.
            pin_memory: Whether to pin memory for the data loaders.
            persistent_workers: Whether to use persistent workers for the data loaders.
            clean_up: Whether to clean up the data after loading.
            offline: Whether to use offline mode.
            save_dir: The directory to save the data.
            verbose: Whether to print verbose output.
            dev_mode: Whether to run in development mode.
            download_only: Whether to only download the data without running an experiment.

        Returns:
            None
        """

        if download_only:
            self.print_master("Download only mode (no experiments)")
            offline = clean_up = validate_contracts = False

            if not modify_contracts:
                self.print_master(
                    "modify_contracts is set to False. You must disable it to activately modify contracts .pkl files."
                )

        action = "Loading" if offline else "Generating"
        self.ctx.log(f"{action} contract datasets")
        (
            self.train_contract_dataset,
            self.val_contract_dataset,
            self.test_contract_dataset,
        ) = get_contract_datasets(
            root=root,
            start_date=start_date,
            end_date=end_date,
            contract_stride=contract_stride,
            interval_min=interval_min,
            right=right,
            target_tte=target_tte,
            tte_tolerance=tte_tolerance,
            moneyness=moneyness,
            strike_band=strike_band,
            volatility_type=volatility_type,
            volatility_scaled=volatility_scaled,
            volatility_scalar=volatility_scalar,
            train_split=train_split,
            val_split=val_split,
            clean_up=clean_up,
            offline=offline,
            save_dir=save_dir,
            verbose=verbose,
            dev_mode=dev_mode,
        )

        self.ctx.log(f"Train contracts: {len(self.train_contract_dataset)}")
        self.ctx.log(f"Validation contracts:{len(self.val_contract_dataset)}")
        self.ctx.log(f"Test contracts:{len(self.test_contract_dataset)}")

        # Warning: this will overwite existing contract .pkl files
        if not modify_contracts:
            self.ctx.log(
                "Safe mode enabled (modify_contracts=False): contracts will not be modified."
            )
        else:
            self.ctx.log(
                "Warning (modify_contracts=True): contracts will be overwritten if invalid."
            )

        if validate_contracts or download_only:
            action = "Validating contracts" if validate_contracts else "Downloading data"
            self.ctx.log(f"{action} with ThetaData API...")
            self.train_contract_dataset = get_forecasting_dataset(
                contract_dataset=self.train_contract_dataset,
                tte_tolerance=tte_tolerance,
                validate_contracts=validate_contracts,
                download_only=download_only,
                modify_contracts=modify_contracts,
                verbose=verbose,
                save_dir=save_dir,
                dev_mode=dev_mode,
            )

            self.val_contract_dataset = get_forecasting_dataset(
                contract_dataset=self.val_contract_dataset,
                tte_tolerance=tte_tolerance,
                validate_contracts=validate_contracts,
                download_only=download_only,
                modify_contracts=modify_contracts,
                verbose=verbose,
                save_dir=save_dir,
                dev_mode=dev_mode,
            )
            self.test_contract_dataset = get_forecasting_dataset(
                contract_dataset=self.test_contract_dataset,
                tte_tolerance=tte_tolerance,
                validate_contracts=validate_contracts,
                download_only=download_only,
                modify_contracts=modify_contracts,
                verbose=verbose,
                save_dir=save_dir,
                dev_mode=dev_mode,
            )

        self.ctx.log(f"Train contracts (validated): {len(self.train_contract_dataset)}")
        self.ctx.log(f"Validation contracts (validated):{len(self.val_contract_dataset)}")
        self.ctx.log(f"Test contracts (validated):{len(self.test_contract_dataset)}")

        if download_only:
            return

        self.train_loader, self.val_loader, self.test_loader, self.scaler = (
            get_forecasting_loaders(
                train_contract_dataset=self.train_contract_dataset,
                val_contract_dataset=self.val_contract_dataset,
                test_contract_dataset=self.test_contract_dataset,
                seq_len=seq_len,
                pred_len=pred_len,
                tte_tolerance=tte_tolerance,
                core_feats=core_feats,
                tte_feats=tte_feats,
                datetime_feats=datetime_feats,
                vol_feats=vol_feats,
                rolling_volatility_range=rolling_volatility_range,
                keep_datetime=keep_datetime,
                target_channels=target_channels,
                target_type=target_type,
                batch_size=batch_size,
                shuffle=shuffle,
                drop_last=drop_last,
                num_workers=num_workers,
                prefetch_factor=prefetch_factor,
                pin_memory=pin_memory,
                persistent_workers=persistent_workers,
                clean_up=clean_up,
                offline=offline,
                save_dir=save_dir,
                verbose=verbose,
                scaling=scaling,
                modify_contracts=modify_contracts,
                intraday=False,
                dtype=dtype,
                dev_mode=dev_mode,
            )
        )

        self.print_master(
            f"PyTorch Dataloaders initialized. Training examples: {len(self.train_loader.dataset)}. Validation examples: {len(self.val_loader.dataset)}. Test examples: {len(self.test_loader.dataset)}."
        )

    def set_seed(self) -> None:
        """Fixes a seed for reproducibility purposes."""
        # Reproducibility
        torch.manual_seed(self.seed)  # CPU
        np.random.seed(self.seed)  # Numpy
        random.seed(self.seed)  # Python
        if torch.cuda.is_available():
            torch.cuda.manual_seed(self.seed)  # GPU
            torch.cuda.manual_seed_all(self.seed)  # multi-GPU

    def init_earlystopping(self, path: str, patience: int) -> None:
        self.early_stopping = EarlyStopping(
            patience=patience,
            path=path,
        )

    def get_sklearn_data(
        self,
        train_loader: Optional[DataLoader] = None,
        val_loader: Optional[DataLoader] = None,
        test_loader: Optional[DataLoader] = None,
        has_datetime: bool = False,
    ) -> None:

        if train_loader is None:
            assert hasattr(
                self, "train_loader"
            ), "A train_loader must be given or initialized using init_loaders() method."
            train_loader = self.train_loader
        if val_loader is None:
            assert hasattr(
                self, "val_loader"
            ), "A val_loader must be given or initialized using init_loaders() method."
            val_loader = self.val_loader
        if test_loader is None:
            assert hasattr(
                self, "test_loader"
            ), "A test_loader must be given or initialized using init_loaders() method."
            test_loader = self.test_loader

        # Iterate through the ConcatDataset list of datasets
        train_x = []; train_y = []; train_dtx = []; train_dty = []
        for dataset in train_loader.dataset.datasets:
            numpy_dataset = dataset.to_numpy()
            train_x.append(numpy_dataset[0]); train_y.append(numpy_dataset[1])
            if has_datetime:
                train_dtx.append(numpy_dataset[2]); train_dty.append(numpy_dataset[3])

        val_x = []; val_y = []; val_dtx = []; val_dty = []
        for dataset in val_loader.dataset.datasets:
            numpy_dataset = dataset.to_numpy()
            val_x.append(numpy_dataset[0]); val_y.append(numpy_dataset[1])
            if has_datetime:
                val_dtx.append(numpy_dataset[2]); val_dty.append(numpy_dataset[3])

        test_x = []; test_y = []; test_dtx = []; test_dty = []
        for dataset in test_loader.dataset.datasets:
            numpy_dataset = dataset.to_numpy()
            test_x.append(numpy_dataset[0]); test_y.append(numpy_dataset[1])
            if has_datetime:
                test_dtx.append(numpy_dataset[2]); test_dty.append(numpy_dataset[3])

        # Concatenate the numpy arrays
        train_x = np.concatenate(train_x + val_x, axis=0)
        train_y = np.concatenate(train_y + val_y, axis=0)
        test_x = np.concatenate(test_x, axis=0)
        test_y = np.concatenate(test_y, axis=0)

        # Flatten the last two dimensions of the input: (num_windows, seq_len, num_features) -> (num_windows, seq_len * num_features)
        if train_x.ndim == 3:
            train_x = train_x.reshape(train_x.shape[0], -1)
            test_x = test_x.reshape(test_x.shape[0], -1)

        # For multi-feature targets, flatten the last two dimensions: (num_windows, pred_len, num_target_features) -> (num_windows, pred_len * num_target_features)
        if train_y.ndim == 3:
            train_y = train_y.reshape(train_y.shape[0], -1)
            test_y = test_y.reshape(test_y.shape[0], -1)

        self.sklearn_data = {
            "train_x": train_x,
            "train_y": train_y,
            "test_x": test_x,
            "test_y": test_y,
        }

        if has_datetime:
            train_dtx = np.concatenate(train_dtx + val_dtx, axis=0)
            train_dty = np.concatenate(train_dty + val_dty, axis=0)
            test_dtx = np.concatenate(test_dtx, axis=0)
            test_dty = np.concatenate(test_dty, axis=0)

            self.sklearn_dt_data = {
                "train_dtx": train_dtx,
                "train_dty": train_dty,
                "test_dtx": test_dtx,
                "test_dty": test_dty,
            }

        return

    def train_sklearn(
        self,
        model: BaseEstimator,
        param_dict: dict,
        tuning_method: str = "grid",
        n_splits: int = 5,
        verbose: int = 1,
        n_jobs: int = -1,
        n_iter: int = 100,
        train_x: Optional[np.ndarray] = None,
        train_y: Optional[np.ndarray] = None,
        target_type: str = "multistep",
        best_model_path: Optional[str] = None,
        early_stopping: bool = False,
        patience: Optional[int] = None,
    ) -> None:

        # Temporally split the data into train and validation sets
        cv = TimeSeriesSplit(n_splits=n_splits)

        # Convert PyTorch DataLoader objects to single NumPy arrays for scikit-learn training
        if all(v is None for v in [train_x, train_y]):

            with self.ctx.status("Converting PyTorch DataLoader to NumPy arrays..."):
                has_datetime = self.train_loader.dataset.datasets[0].has_datetime # Check if datetime is on for the first ForecastingDataset
                self.get_sklearn_data(
                    train_loader=self.train_loader,
                    val_loader=self.val_loader,
                    test_loader=self.test_loader,
                    has_datetime=has_datetime,
                )
                train_x = self.sklearn_data["train_x"]
                train_y = self.sklearn_data["train_y"]

            self.ctx.log(f"train_x shape: {train_x.shape}")
            self.ctx.log(f"train_y shape: {train_y.shape}")

            # Check if NaNs in train_x or train_y
            if np.isnan(train_x).any():
                # Compute how many NaNs there are
                num_nans = np.isnan(train_x).sum()
                self.ctx.log(f"train_x contains {num_nans} NaN values.")

            if np.isnan(train_y).any():
                num_nans = np.isnan(train_y).sum()
                self.ctx.log(f"train_y contains {num_nans} NaN values.")

            if np.isnan(train_x).any() or np.isnan(train_y).any():
                raise ValueError("train_x or train_y contains NaN values.")

        if target_type in ["multistep", "average"]:
            scoring = "neg_root_mean_squared_error"
        elif target_type == "average_direction":
            scoring = "accuracy"
        else:
            raise ValueError(f"Invalid target type: {target_type}")

        if tuning_method == "grid":
            clf = GridSearchCV(
                estimator=model,
                param_grid=param_dict,
                scoring=scoring,
                n_jobs=n_jobs,
                cv=cv,
                verbose=verbose,
            )
        elif tuning_method == "random":
            clf = RandomizedSearchCV(
                estimator=model,
                param_distributions=param_dict,
                scoring=scoring,
                n_jobs=n_jobs,
                cv=cv,
                verbose=verbose,
                n_iter=n_iter,
            )
        else:
            raise ValueError(f"Invalid tuning method: {tuning_method}")


        # Print out model param_dict with Rich table
        table = Table(title="Hyperparameter Grid")
        table.add_column("Parameter", style="bold cyan")
        table.add_column("Values", style="magenta")
        for k, v in param_dict.items():
            table.add_row(str(k), str(v))
        self.ctx.log(table)

        with self.ctx.status("Fitting sklearn model..."):
            search = clf.fit(X=train_x, y=train_y)

        # Save model
        best_model = search.best_estimator_
        if best_model_path is None:
            self.best_model_path = self.log_dir / "model.pkl"
        else:
            self.best_model_path = Path(best_model_path)

        # Create directory if it doesn't exist
        self.best_model_path.parent.mkdir(parents=True, exist_ok=True)
        joblib.dump(best_model, self.best_model_path) # Save the best model

        # TODO: Add model.upload() for neptune logging
        self.print_master(f"Best model saved to {self.best_model_path}")

        # Log best model parameters
        self.logger["best_model/params"] = search.best_params_
        self.logger["best_model/score"] = search.best_score_

    def test_sklearn(
        self,
        metrics: List[str],
        target_type: str = "multistep",
        test_x: Optional[np.ndarray] = None,
        test_y: Optional[np.ndarray] = None,
        best_model: Optional[BaseEstimator] = None,
    ) -> None:

        if best_model is None:
            best_model = joblib.load(self.best_model_path)

        # Evaluate best model on the test set
        if all(v is None for v in [test_x, test_y]):
            test_x = self.sklearn_data["test_x"]
            test_y = self.sklearn_data["test_y"]

        test_preds = best_model.predict(X=test_x)
        test_metrics, metric_keys = get_metrics(
            target=test_y,
            output=test_preds,
            metrics=metrics,
            target_type=target_type,
        )
        train_preds = best_model.predict(X=self.sklearn_data["train_x"])
        train_metrics, _ = get_metrics(
            target=self.sklearn_data["train_y"],
            output=train_preds,
            metrics=metrics,
            target_type=target_type,
        )

        # Return metrics in dictionary format
        test_stats = dict()
        train_stats = dict()
        for i, metric in enumerate(metric_keys):
            test_stats[metric] = test_metrics[i]
            train_stats[metric] = train_metrics[i]

        self.log_stats(stats=test_stats, metrics=metrics, mode="test")
        self.log_stats(stats=train_stats, metrics=metrics, mode="train")

    def train_torch(
        self,
        model: Union[nn.Module, BaseEstimator],
        optimizer: optim.Optimizer,
        criterion: nn.Module,
        num_epochs: int,
        device=None,
        train_loader: Optional[DataLoader] = None,
        val_loader: Optional[DataLoader] = None,
        metrics: List[str] = ["mse"],
        best_model_metric: str = "mse",
        best_model_path: Optional[str] = None,
        early_stopping: bool = False,
        patience: Optional[int] = None,
        scheduler: Optional[_LRScheduler] = None,
        target_type: str = "multistep",
    ) -> nn.Module:
        """
            Trains a model.

        Args:
            model: The model to train.
            optimizer: The optimizer to use.
            criterion: The loss function.
            train_loader: The training data loader.
            num_epochs: The number of epochs to train.
            val_loader: The validation data loader.
            metrics: The evaluation metrics to track.
            best_model_path: The path to save the best model.
            early_stopping: Whether to use early stopping.
            patience: The number of epochs to wait before stopping.
            scheduler: The learning rate scheduler.
        """

        if device is None:
            device = self.device

        if train_loader is None:
            assert hasattr(
                self, "train_loader"
            ), "A train_loader must be given or initialized using init_loaders() method."
            train_loader = self.train_loader

        if val_loader is None:
            assert hasattr(
                self, "val_loader"
            ), "A val_loader must be given or initialized using init_loaders() method."
            val_loader = self.val_loader

        if best_model_path is None:
            self.best_model_path = self.log_dir / "model.pth"
        else:
            self.best_model_path = best_model_path

        # Deep learning (PyTorch) pipeline
        num_examples = len(train_loader.dataset)
        self.print_master(f"Training on {num_examples} examples")

        self.best_val_metric = float("inf")
        if early_stopping:
            assert (
                patience is not None
            ), "Patience must be specified for early stopping."
            self.init_earlystopping(patience=patience, path=self.best_model_path)
            self.print_master("Early stopping initialized")

        # <--------------- Training --------------->
        for epoch in range(num_epochs):
            model.train()
            total_loss = torch.tensor(0.0, device=device)
            running_loss = torch.tensor(0.0, device=device)
            running_num_examples = torch.tensor(0.0, device=device)
            start_time = time.time()

            for i, batch in enumerate(train_loader):
                x = batch[0].to(device)
                y = batch[1].to(device)
                optimizer.zero_grad()
                output = model(x)
                loss = criterion(output, y)

                # Metrics
                num_batch_examples = torch.tensor(batch[0].shape[0], device=device)
                total_loss += loss * num_batch_examples
                running_loss += loss * num_batch_examples
                running_num_examples += num_batch_examples

                # Update model parameters
                loss.backward()
                optimizer.step()

                # Periodic Logging
                if (i + 1) % 100 == 0:

                    # Loss metrics
                    loss_tensor = running_loss.to(device)
                    num_examples_tensor = running_num_examples.to(device)
                    end_time = time.time()

                    # Only rank 0 prints and logs details
                    average_loss = loss_tensor.item() / num_examples_tensor.item()
                    self.print_master(
                        f"[Epoch {epoch}, Batch ({i+1}/{len(train_loader)})]: {end_time - start_time:.3f}s. Loss: {average_loss:.6f}"
                    )

                    # Reset trackers
                    running_loss = torch.tensor(0.0, device=device)
                    running_num_examples = torch.tensor(0.0, device=device)
                    start_time = time.time()

                if scheduler:
                    scheduler.step()

            # Average Loss + Logging
            epoch_loss = total_loss.item() / num_examples
            self.print_master(f"Epoch {epoch}. Training loss: {epoch_loss:.6f}.")
            self.epoch_logger(self.logger, "train/loss", epoch_loss)

            # <--------------- Validation --------------->
            if val_loader:
                self.validate_torch(
                    model=model,
                    val_loader=val_loader,
                    criterion=criterion,
                    metrics=metrics,
                    epoch=epoch,
                    best_model_metric=best_model_metric,
                    best_model_path=self.best_model_path,
                    early_stopping=early_stopping,
                    device=device,
                    target_type=target_type,
                )

            # Early stopping
            if early_stopping and self.early_stopping.early_stop:
                self.print_master("EarlyStopping activated, ending training.")
                break

            # Checkpoint (online)
            if self.logging == "neptune":
                pass
            else:
                run_time = time.time() - self.start_time
                self.logger["parameters/running_time"] = format_time_dynamic(run_time)

        # Upload best model to Neptune
        if self.logging == "neptune":
            self.print_master("Uploading best model to Neptune.")
            self.logger[f"model_checkpoints/{self.exp_id}"].upload(best_model_path)

        return model

    def validate_torch(
        self,
        model: nn.Module,
        val_loader: DataLoader,
        criterion: nn.Module,
        device,
        best_model_metric: str = "mse",
        metrics: List[str] = ["mse"],
        epoch: Optional[int] = None,
        best_model_path: Optional[str] = None,
        early_stopping: Optional[bool] = False,
        target_type: str = "multistep",
    ) -> None:
        """
        Validates the model.

        Args:
            model: The model to validate.
            val_loader: The validation data loader.
            criterion: The loss function.
            best_model_metric: The metric to use for the best model.
            metrics: The evaluation metrics to track.
            epoch: The current epoch.
            best_model_path: The path to save the best model.
            early_stopping: Whether to use early stopping

        Returns:
            None

        """
        stats = self.evaluate_torch(
            model=model,
            loader=val_loader,
            criterion=criterion,
            metrics=metrics,
            device=device,
            target_type=target_type,
        )
        self.log_stats(stats=stats, metrics=metrics, mode="val")

        if best_model_metric in {"mse", "mae", "rmse", "mape"}:
            val_metric = stats[best_model_metric] # Minimizing metrics
        elif best_model_metric in {"r2", "accuracy", "precision", "recall", "f1", "auc"}:
            val_metric = -stats[best_model_metric] # Maximizing metrics
        else:
            raise ValueError(f"Invalid best model metric: {best_model_metric}")

        # Save best model and apply early stopping
        if early_stopping:
            self.early_stopping(val_metric, model)
        else:
            if val_metric < self.best_val_metric:
                self.print_master(
                    f"Validation {best_model_metric} decreased ({self.best_val_metric:.6f} --> {val_metric:.6f})."
                )
                path_dir = Path(best_model_path).parent.absolute()
                if not path_dir.is_dir():
                    path_dir.mkdir(parents=True)
                torch.save(model.state_dict(), best_model_path)
                self.print_master(f"Saving Model Weights at: {best_model_path}...")
                self.best_val_metric = val_metric

        self.print_master("Validation complete")

    def test_torch(
        self,
        model: Union[nn.Module, BaseEstimator],
        criterion: nn.Module,
        device=None,
        test_loader: Optional[DataLoader] = None,
        metrics: List[str] = ["mse"],
        target_type: str = "multistep",
    ) -> None:
        """

        Tests the model.

        Args:
            model: The model to test.
            test_loader: The test data loader.
            criterion: The loss function.
            metrics: The evaluation metrics to track.

        Returns:
            None

        """
        if device is None:
            device = self.device

        if test_loader is None:
            assert hasattr(
                self, "test_loader"
            ), "A test_loader must be given or initialized using init_loaders() method."
            test_loader = self.test_loader

        # Load best model
        model_weights = torch.load(self.best_model_path)
        model.load_state_dict(model_weights)

        # Test set evaluation
        stats = self.evaluate_torch(
            model=model,
            loader=test_loader,
            criterion=criterion,
            metrics=metrics,
            device=device,
            target_type=target_type,
        )

        self.log_stats(stats=stats, metrics=metrics, mode="test")

    def evaluate(
        self,
        model: Union[nn.Module, BaseEstimator],
        loader: DataLoader,
        criterion: nn.Module,
        device,
        metrics: List[str] = ["mse"],
        target_type: str = "multistep",
    ) -> dict:
        """
        Evaluates the model and returns metrics. Used in validation and testing.

        Args:
            model: The model to evaluate.
            loader: The data loader.
            criterion: The loss function.
            metrics: The evaluation metrics to track.

        Returns:
            stats: A dictionary of evaluation metrics.
        """
        self.print_master("Evaluating...")
        stats = dict()
        num_examples = len(loader.dataset)  # Total number of examples across all ranks

        model.eval()
        with torch.no_grad():

            # Initialize Metrics
            total_metrics = np.zeros(len(metrics))

            for i, batch in enumerate(loader):
                x = batch[0].to(device)
                y = batch[1].to(device)
                output = model(x)

                # Compute metrics
                num_batch_examples = batch[0].shape[0]
                batch_metrics, metric_keys = get_metrics(
                    metrics=metrics,
                    output=output,
                    target=y,
                    target_type=target_type,
                )
                total_metrics +=  batch_metrics * num_batch_examples

        # Return metrics in dictionary format
        stats = dict()
        for i, metric in enumerate(metric_keys):
            stats[metric] = total_metrics[i] / num_examples

        return stats

    def log_universe(
        self,
        market_metrics,
        root: str,
        parent: str,
    ) -> None:
        """
        Logs the universe to the logger.

        Args:
            root: The root directory containing the data.
            universe: The Universe object.
            parent: The parent experiment ID.

        Returns:
            None
        """
        pass

        # Find the parent experiment file
        # market_metrics =

        # if self.logging == "neptune":
        #     self.logger["universe"].upload(universe)
        # elif self.logging == "offline":
        #     with open(self.log_file, "w") as f:
        #         json.dump(universe, f, indent=2)
        # else:
        #     raise ValueError(f"Invalid logging method: {self.logging}.")

    def log_stats(self, stats: dict, metrics: List[str], mode: str):
        modes = {"val": "Validation", "test": "Test", "train": "Train"}
        Mode = modes[mode]

        for metric in metrics:
            self.print_master(f"{Mode} {metric}: {stats[metric]:.6f}")
            self.epoch_logger(self.logger, f"{mode}/{metric}", stats[metric])

    def epoch_logger(self, logger, key: str, value: str) -> None:
        if self.logging == "neptune":
            logger[key].append(value)
        elif self.logging == "offline":
            if key not in logger:
                logger[key] = []
            logger[key].append(value)
        else:
            raise ValueError(f"Invalid logging method: {self.logging}.")

    def print_master(self, message: str):
        """
        Prints statements to the rank 0 node.
        """
        self.ctx.log(message)



from copy import deepcopy
from pathlib import Path
from typing import Tuple, List, Union, Optional
from rich.console import Console
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler
import torch
from torch.utils.data import DataLoader, Dataset, ConcatDataset
from datetime import datetime, timedelta
import pandas_market_calendars as mcal


# Datasets
from optrade.data.thetadata import find_optimal_strike, get_expirations
from optrade.data.contracts import Contract, ContractDataset
from optrade.utils.error_handlers import (
    DataValidationError,
    INCOMPATIBLE_START_DATE,
    INCOMPATIBLE_END_DATE,
    NAN_FEATURES,
)

# Features
from optrade.data.features import transform_features
from optrade.utils.misc import datetime_to_tensor, tensor_to_datetime

# Get absolute path for this script
SCRIPT_DIR = Path(__file__).resolve().parent


class ForecastingDataset(Dataset):
    def __init__(
        self,
        data: pd.DataFrame,
        seq_len: int,
        pred_len: int,
        target_channels: Optional[List[str]] = None,
        target_type: str = "multistep",
        dtype: str = "float32",
    ) -> None:
        """
        Initializes the ForecastingDataset class.

        Args:
            data (pd.DataFrame):
                Input DataFrame containing the time series data.
            seq_len (int):
                Length of the lookback window for each sample.
            pred_len (int):
                Length of the forecast window (number of steps ahead to predict).
            target_channels (Optional[List[str]]):
                List of column names to include as target channels. If None, all columns are used.
            target_type (str):
                Type of target to predict. Must be one of:
                - "multistep": Predicts the full future sequence (regression).
                - "average": Predicts the average value over the forecast window (regression).
                - "average_direction": Predicts the sign of the average change (binary classification).
            dtype (str):
                Data type for the internal PyTorch tensors (e.g., "float32", "float64"). Default is "float32".

        Returns:
            None
        """

        # Numeric data conversion (remove datetime strings from direct features)
        self.has_datetime = "datetime" in data.columns
        if self.has_datetime:
            self.datetime = data["datetime"].values  # Store as numpy array
            data_numeric = data.drop(columns=["datetime"]).to_numpy()
        else:
            data_numeric = data.to_numpy()

        # Check for NaNs using the original DataFrame (with column names)
        columns = (
            data.drop(columns=["datetime"]).columns
            if self.has_datetime
            else data.columns
        )
        nan_counts = data[columns].isna().sum()

        if nan_counts.any():
            bad_feats = nan_counts[nan_counts > 0]
            message = (
                f"[ForecastingDataset] Found NaNs in {len(bad_feats)} feature(s):\n"
                f"{bad_feats.to_string()}\n"
                "This is often caused by illiquid instruments (e.g., no volume or quote data) "
                "or unstable features like LOB imbalance or spread when interval_min is too low.\n"
                "Consider dropping these features or resampling to a longer interval by increasing interval_min parmaeter."
            )
            raise DataValidationError(
                message=message, error_code=NAN_FEATURES, verbose=True, warning=False
            )

        self.torch_dtype = eval("torch." + dtype)
        self.np_dtype = eval("np." + dtype)
        self.data = torch.tensor(data_numeric, dtype=self.torch_dtype)
        self.seq_len = seq_len
        self.pred_len = pred_len
        self.target_type = target_type

        if target_channels is not None and len(target_channels) > 0:
            feature_names = data.columns.to_list()
            self.target_channels_idx = [
                feature_names.index(channel) for channel in target_channels
            ]

    def __len__(self) -> int:
        """
        Returns the number of input-target pairs in the dataset.
        """
        return self.data.shape[0] - self.seq_len - self.pred_len

    def __getitem__(self, idx: int) -> Union[
        Tuple[torch.Tensor, torch.Tensor],
        Tuple[torch.Tensor, torch.Tensor, np.ndarray, np.ndarray],
    ]:
        """Get a sample from the dataset.

        This method retrieves an input-target pair at the specified index, with input being
        the lookback window and target being the forecast window based on the target_type.

        Args:
            idx: Index of the starting point of the lookback window.

        Returns:
            If datetime is available:
                tuple: A tuple containing (input_tensor, target_tensor, input_datetime, target_datetime)
                    - input_tensor: Lookback window of shape (num_features, seq_len).
                    - target_tensor: Target window with shape depending on target_type:
                      - "multistep": (num_target_features, pred_len)
                      - "average": (num_target_features, 1)
                      - "average_direction": (num_target_features, 1)
                    - input_datetime: Datetime values for input window of shape (seq_len,).
                    - target_datetime: Datetime values for target window of shape (pred_len,).
            Otherwise:
                tuple: A tuple containing (input_tensor, target_tensor)
                    - input_tensor: Lookback window of shape (num_features, seq_len).
                    - target_tensor: Target window with shape as described above depending on `target_type` arg.
        """
        input = self.data[idx : idx + self.seq_len]

        if hasattr(self, "target_channels_idx"):
            target = self.data[
                idx + self.seq_len : idx + self.seq_len + self.pred_len,
                self.target_channels_idx,
            ]
        else:
            target = self.data[idx + self.seq_len : idx + self.seq_len + self.pred_len]

        input_tensor = input.transpose(0, 1)
        target_tensor = target.transpose(0, 1)

        if self.target_type == "average":
            target_tensor = target_tensor.mean(dim=0).unsqueeze(0)
        elif self.target_type == "average_direction":
            target_tensor = (target_tensor.mean(dim=0) > 0).unsqueeze(0).float()
        elif self.target_type == "multistep":
            pass
        else:
            raise ValueError(
                "Invalid target_type. Options: 'multistep', 'average', or 'average_direction'."
            )

        if self.has_datetime:
            input_datetime = self.datetime[idx : idx + self.seq_len]
            target_datetime = self.datetime[
                idx + self.seq_len : idx + self.seq_len + self.pred_len
            ]

            # Convert datetime arrays to tensors
            input_datetime_tensor = datetime_to_tensor(input_datetime)
            target_datetime_tensor = datetime_to_tensor(target_datetime)

            return (
                input_tensor,
                target_tensor,
                input_datetime_tensor,
                target_datetime_tensor,
            )
        else:
            return input_tensor, target_tensor

    def to_numpy(self) -> Union[
        Tuple[np.ndarray, np.ndarray],
        Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray],
    ]:
        """Converts the dataset into a set of NumPy arrays for scikit-learn model training.
        Returns:
            Tuple[np.ndarray, np.ndarray]: A tuple containing:
                - inputs: NumPy array of shape (num_samples, seq_len, num_features).
                - targets: NumPy array of shape (num_samples, pred_len, num_target_features).
            If datetime is available:
                - input_datetimes: NumPy array of shape (num_samples, seq_len).
                - target_datetimes: NumPy array of shape (num_samples, pred_len).
        """
        if self.target_type == "multistep":
            num_target_features = self.data.shape[1]
            target_dtype = self.np_dtype
        elif self.target_type == "average":
            num_target_features = 1
            target_dtype = self.np_dtype
        elif self.target_type == "average_direction":
            num_target_features = 1
            target_dtype = np.float32
        else:
            raise ValueError(
                "Invalid target_type. Options: 'multistep', 'average', or 'average_direction'."
            )
        inputs = np.empty(
            (len(self), self.seq_len, self.data.shape[1]), dtype=self.np_dtype
        )
        targets = np.empty(
            (len(self), self.pred_len, num_target_features), dtype=target_dtype
        )

        if self.has_datetime:
            # Specify the unit here to match what tensor_to_datetime uses
            unit = "s"  # seconds
            # Create the arrays with explicit time unit
            input_datetimes = np.empty(
                (len(self), self.seq_len), dtype=f"datetime64[{unit}]"
            )
            target_datetimes = np.empty(
                (len(self), self.pred_len), dtype=f"datetime64[{unit}]"
            )

        for i in range(len(self)):
            item = self.__getitem__(i)
            input_tensor, target_tensor = item[0], item[1]
            inputs[i] = input_tensor.numpy().T
            targets[i] = target_tensor.numpy().T

            if self.has_datetime:
                input_datetime_tensor, target_datetime_tensor = item[2], item[3]
                input_datetimes[i] = tensor_to_datetime(
                    input_datetime_tensor, batch_mode=False
                )
                target_datetimes[i] = tensor_to_datetime(
                    target_datetime_tensor, batch_mode=False
                )

        if self.has_datetime:
            return inputs, targets, input_datetimes, target_datetimes
        else:
            return inputs, targets

    def get_item(self, idx: int) -> Union[
        Tuple[torch.Tensor, torch.Tensor],
        Tuple[torch.Tensor, torch.Tensor, np.ndarray, np.ndarray],
    ]:
        """Get a sample from the dataset.
        This method retrieves an input-target pair at the specified index, with input being
        the lookback window and target being the forecast window based on the target_type.
        Args:
            idx: Index of the starting point of the lookback window.
        Returns:
            If datetime is available:
                tuple: A tuple containing (input_tensor, target_tensor, input_datetime, target_datetime)
                    - input_tensor: Lookback window of shape (num_features, seq_len).
                    - target_tensor: Target window with shape depending on target_type:
                      - "multistep": (num_target_features, pred_len)
                      - "average": (num_target_features, 1)
                      - "average_direction": (num_target_features, 1)
                    - input_datetime: Datetime values for input window of shape (seq_len,).
                    - target_datetime: Datetime values for target window of shape (pred_len,).
            Otherwise:
                tuple: A tuple containing (input_tensor, target_tensor)
                    - input_tensor: Lookback window of shape (num_features, seq_len).
                    - target_tensor: Target window with shape as described above.
        """
        return self.__getitem__(idx)


def normalize_concat_dataset(
    concat_dataset: ConcatDataset,
    scaler: StandardScaler,
) -> None:
    """
    Modifies the data in a ConcatDataset in-place by normalizing it using a fitted StandardScaler.

    Args:
        concat_dataset: ConcatDataset object containing ForecastingDatasets
        scaler: Fitted StandardScaler from scikit-learn.

    Returns:
        None
    """
    for dataset in concat_dataset.datasets:
        data = dataset.data.numpy()

        # Normalize data
        normalized_data = scaler.transform(data)

        # Replace data with normalized version
        dataset.data = torch.tensor(normalized_data, dtype=dataset.torch_dtype)


def normalize_datasets(
    train_dataset: ConcatDataset,
    val_dataset: ConcatDataset,
    test_dataset: ConcatDataset,
) -> Tuple[ConcatDataset, ConcatDataset, ConcatDataset, StandardScaler]:
    """
    Normalizes financial time series datasets using StandardScaler.
    Fits scaler only on training data to prevent look-ahead bias.

    Args:
        train_dataset: Training dataset (ConcatDataset of ForecastingDatasets)
        val_dataset: Validation dataset
        test_dataset: Test dataset

    Returns:
        Tuple[ConcatDataset, ConcatDataset, ConcatDataset, StandardScaler]: Normalized training, validation, and test datasets, and the fitted Standard
    """
    # Extract all underlying data from train_dataset
    all_train_data = []

    # Iterate through the individual ForecastingDataset objects within the ConcatDataset
    for dataset_idx in range(len(train_dataset.datasets)):
        individual_dataset = train_dataset.datasets[dataset_idx]

        # Extract data tensor from the individual dataset
        data = individual_dataset.data.numpy()
        all_train_data.append(data)

    # Stack all the arrays together
    combined_train_data = np.vstack(all_train_data)

    # Fit scaler on training data only
    scaler = StandardScaler()
    scaler.fit(combined_train_data)

    # Apply normalization to all datasets
    normalize_concat_dataset(train_dataset, scaler)
    normalize_concat_dataset(val_dataset, scaler)
    normalize_concat_dataset(test_dataset, scaler)

    return train_dataset, val_dataset, test_dataset, scaler


def get_forecasting_dataset(
    contract_dataset: ContractDataset,
    tte_tolerance: Tuple[int, int],
    seq_len: Optional[int] = None,
    pred_len: Optional[int] = None,
    core_feats: List[str] = ["option_returns"],
    tte_feats: Optional[List[str]] = None,
    datetime_feats: Optional[List[str]] = None,
    vol_feats: Optional[List[str]] = None,
    rolling_volatility_range: Optional[List[int]] = None,
    keep_datetime: bool = False,
    target_type: str = "multistep",
    clean_up: bool = False,
    offline: bool = False,
    intraday: bool = False,
    target_channels: Optional[List[str]] = None,
    dtype: str = "float32",
    save_dir: Optional[str] = None,
    download_only: bool = False,
    validate_contracts: bool = False,
    modify_contracts: bool = False,
    verbose: bool = False,
    warning: bool = True,
    dev_mode: bool = False,
) -> Union[ContractDataset, Tuple[ConcatDataset, ContractDataset]]:
    """
    Creates a PyTorch dataset object composed of multiple ForecastingDatasets, each representing
    different option contracts.

    Args:
        contract_dataset: ContractDataset object containing option contract parameters
        tte_tolerance: Tuple of (min, max) time to expiration tolerance in days
        core_feats: List of core features to include
        tte_feats: List of time-to-expiration features to include
        datetime_feats: List of datetime features to include
        vol_feats: List of volatility features to include
        rolling_volatility_range: List of rolling volatility ranges to include
        keep_datetime: Whether to keep the datetime column in the dataset
        target_type: Type of forecasting target. Options: "multistep" (float), "average" (float), or "average_direction" (binary).
        clean_up: Whether to clean up the data after use
        offline: Whether to load saved contracts from disk
        intraday: Whether to use intraday data
        target_channels: List of target channels to include in the target tensor. If None, all channels will be included.
        seq_len: Sequence length of lookback window (input)
        pred_len: Prediction length of forecast window (target)
        dtype: Data type for the PyTorch tensors
        save_dir: Save directory
        download_only: Whether to download data only (used mainly for Universe class)
        validate_contracts: Whether to validate contracts by requesting data from ThetaData API and adjustintg start and end dates if necessary.
        modify_contracts: Whether to delete old contracts .pkl file and save the (new) validate contracts in the same path. Warning: This will overwrite the old contracts.
        verbose: Whether to print verbose output
        warning: Whether to print verbose DataValidationError statements as warnings or errors.
        dev_mode: Whether to run in development mode.

    Returns:
        ContractDataset: The updated ContractDataset object if `download_only`=True or `validate_contracts`=True.
        Tuple[ConcatDataset, ContractDataset]: A tuple containing the concatenated PyTorch dataset and the updated ContractDataset if download_only=False.
    """

    ctx = Console()
    dataset_list = []
    validated_contracts = []  # Track which contracts are actually valid
    updated_contracts = []  # Track contracts that were updated (not in initial list)
    expirations_exist = False
    tried_contracts = set()

    # First, validate that download_only and validate_contracts aren't both True
    if download_only and validate_contracts:
        raise ValueError(
            "Please use download_only and validate_contracts separately. Both cannot be True."
        )

    if download_only:
        clean_up = False
        offline = False
    elif validate_contracts:
        clean_up = True
        offline = False
    else:
        assert seq_len is not None, "seq_len must be provided for forecasting dataset"
        assert pred_len is not None, "pred_len must be provided for forecasting dataset"

    # Save initial contracts to compare adjusted contracts later
    initial_contracts = deepcopy(contract_dataset.contracts)

    # Iterate through each contract in the ContractDataset
    for (
        contract
    ) in (
        initial_contracts
    ):  # Use initial_contracts instead to avoid modifying during iteration
        # Flag to track if we should try the next contract
        move_to_next_contract = False
        original_contract = deepcopy(contract)  # Keep original for comparison

        # Continue trying the current contract until we succeed or decide to move on
        while not move_to_next_contract:
            try:
                df = contract.load_data(
                    save_dir=save_dir,
                    clean_up=clean_up,
                    offline=offline,
                    warning=warning,
                    dev_mode=dev_mode,
                )

                if not (download_only or validate_contracts):
                    # Select and add features
                    data = transform_features(
                        df=df,
                        core_feats=core_feats,
                        tte_feats=tte_feats,
                        datetime_feats=datetime_feats,
                        vol_feats=vol_feats,
                        rolling_volatility_range=rolling_volatility_range,
                        root=contract.root,
                        right=contract.right,
                        strike=contract.strike,
                        exp=contract.exp,
                        keep_datetime=keep_datetime,
                    )

                    # Convert to PyTorch dataset
                    dataset = ForecastingDataset(
                        data=data,
                        seq_len=seq_len,
                        pred_len=pred_len,
                        target_channels=target_channels,
                        target_type=target_type,
                        dtype=dtype,
                    )
                    dataset_list.append(dataset)

                # If contract was modified from its original state, add to updated contracts
                if contract != original_contract:
                    updated_contracts.append(contract)

                # Add valid contract to validated list
                validated_contracts.append(contract)
                move_to_next_contract = True

            except DataValidationError as e:
                # Handle date-related errors that might be recoverable
                if e.error_code in [INCOMPATIBLE_START_DATE, INCOMPATIBLE_END_DATE]:
                    candidate_start_date = e.real_start_date
                    candidate_exp = (
                        e.real_end_date
                        if e.error_code == INCOMPATIBLE_END_DATE
                        else contract.exp
                    )

                    move_to_next_contract, new_contract = calibrate_new_contract(
                        contract_dataset=contract_dataset,
                        original_contract=original_contract,
                        candidate_start_date=candidate_start_date,
                        candidate_exp=candidate_exp,
                        tte_tolerance=tte_tolerance,
                        expirations_exist=expirations_exist,
                        save_dir=save_dir,
                        verbose=verbose,
                        dev_mode=dev_mode,
                    )
                    expirations_exist = True


                    # If calibration suggests retrying with a new contract...
                    if not move_to_next_contract and new_contract is not None:
                        contract_key = (new_contract.start_date, new_contract.exp)

                        if contract_key in tried_contracts:
                            if verbose:
                                ctx.log(f"Previously tried contract {contract_key} failed. Skipping.")
                            move_to_next_contract = True  # Skip this contract
                            continue

                        tried_contracts.add(contract_key)
                        contract = new_contract  # Try the calibrated contract

                elif e.error_code == NAN_FEATURES:
                    raise e  # Raise the error to be handled outside
                else:
                    if verbose:
                        ctx.log(
                            f"DataValidationError for {contract}: {e}. Moving to next contract."
                        )
                    move_to_next_contract = True

            except Exception as e:
                if verbose:
                    ctx.log(
                        f"Unknown error for {contract}: {e}. Moving to next contract."
                    )
                raise e
                # move_to_next_contract = True

    # Find removed contracts (in initial but not in validated)
    removed_contracts = [c for c in initial_contracts if c not in validated_contracts]

    # Find truly new contracts (in updated but not equivalent to any in initial)
    # This is different from just comparing sets because contracts might be updated versions
    new_contracts = [c for c in updated_contracts]

    # Display changes if verbose
    if verbose and (removed_contracts or new_contracts):
        from rich.table import Table

        # Create a table for removed contracts
        if removed_contracts:
            removed_table = Table(title="Invalid Contracts", show_header=True)
            removed_table.add_column("Root", style="cyan")
            removed_table.add_column("Start Date", style="cyan")
            removed_table.add_column("Expiration", style="cyan")
            removed_table.add_column("Strike", style="cyan")
            removed_table.add_column("Right", style="cyan")

            for contract in removed_contracts:
                removed_table.add_row(
                    contract.root,
                    contract.start_date,
                    contract.exp,
                    str(contract.strike),
                    contract.right,
                    style="red",
                )
            ctx.print(removed_table)

        # Create a table for updated/new contracts
        if updated_contracts:
            updated_table = Table(title="Updated (New Contracts", show_header=True)
            updated_table.add_column("Root", style="cyan")
            updated_table.add_column("Start Date", style="cyan")
            updated_table.add_column("Expiration", style="cyan")
            updated_table.add_column("Strike", style="cyan")
            updated_table.add_column("Right", style="cyan")
            updated_table.add_column("Status", style="cyan")

            for contract in updated_contracts:
                # Check if this is a modified version of an original contract
                is_update = False
                for orig in initial_contracts:
                    if (
                        contract.root == orig.root
                        and contract.strike == orig.strike
                        and contract.interval_min == orig.interval_min
                        and contract.right == orig.right
                    ):
                        is_update = True
                        break

                status = "Updated" if is_update else "New"

                updated_table.add_row(
                    contract.root,
                    contract.start_date,
                    contract.exp,
                    str(contract.strike),
                    contract.right,
                    status,
                    style="green",
                )
            ctx.print(updated_table)

    # Replace the contract list with only validated contracts
    contract_dataset.contracts = validated_contracts

    # Save contract changes if needed
    contracts_changed = set(validated_contracts) != set(initial_contracts)
    if contracts_changed and modify_contracts:
        if verbose:
            ctx.log(
                f"Contract changes detected: {len(removed_contracts)} removed, {len(updated_contracts)} updated/added"
            )
            ctx.log("Saving and overwriting ContractDataset .pkl file")
        contract_dataset.save(clean_file=True)

    if download_only or validate_contracts:
        return contract_dataset
    else:
        if len(dataset_list) == 0:
            raise ValueError(
                "No valid contracts found. All contracts were invalid or contained errors."
            )
        return ConcatDataset(dataset_list), contract_dataset


def calibrate_new_contract(
    contract_dataset: ContractDataset,
    original_contract: Contract,
    candidate_start_date: str,
    candidate_exp: str,
    tte_tolerance: Tuple[int, int],
    expirations_exist: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
    dev_mode: bool = False,
) -> Tuple[bool, Optional[Contract]]:

    # Rich console
    ctx = Console()

    # <------Calibrate New Expiration------>
    if verbose:
        ctx.log(
            f"Calibrating new expiration date '{candidate_exp}' for original contract..."
        )

    # Query list of expirations to find closests expiration to real_end_date
    expirations_offline = (
        True if expirations_exist else False
    )  # Do not query API if expirations.csv exists
    expirations = get_expirations(
        root=original_contract.root,
        save_dir=save_dir,
        clean_up=False,
        offline=expirations_offline,
        dev_mode=dev_mode,
    )
    expirations["date"] = pd.to_datetime(
        expirations["date"].astype(str), format="%Y%m%d"
    )  # Ensure dates are in datetime format
    candidate_exp = pd.to_datetime(
        candidate_exp, format="%Y%m%d"
    )  # Ensure candidate expiration is in datetime format
    closest_date = (expirations["date"] - candidate_exp).abs().idxmin()
    new_exp = expirations.loc[closest_date, "date"].strftime("%Y%m%d")

    if verbose:
        ctx.log(f"Closest expiration date found: {new_exp}")

    # <------Calibrate New Start Date------>
    if verbose:
        ctx.log(
            f"Calibrating new start date '{candidate_start_date}' for original contract..."
        )
    # Validate that the new start date doesn't fall on a weekend or market holiday, otherwise move forward in time to next valid date
    new_start_date = get_valid_start_date(candidate_start_date)

    if verbose:
        ctx.log(f"New start date found: {new_start_date}")

    # <------TTE Tolerance Check------>
    # Check if the timespan from new_start_date to new_exp is within time-to-expiration tolerance
    start_date_dt = pd.to_datetime(new_start_date, format="%Y%m%d")
    exp_date_dt = pd.to_datetime(new_exp, format="%Y%m%d")
    time_span = exp_date_dt - start_date_dt

    if time_span >= pd.Timedelta(days=tte_tolerance[0]):
        # Update the contract if it meets the tolerance
        if verbose:
            ctx.log(
                f"TTE Tolerance = {time_span} check passed: updating contract with new start date: {new_start_date} and expiration: {new_exp}"
            )

        # Create a new contract and modify parameters
        contract = deepcopy(original_contract)
        contract.start_date = new_start_date
        contract.exp = new_exp

        # Update strike based on new parameters
        contract.strike = find_optimal_strike(
            root=contract.root,
            start_date=new_start_date,
            exp=new_exp,
            right=contract_dataset.right,
            interval_min=contract_dataset.interval_min,
            moneyness=contract_dataset.moneyness,
            strike_band=contract_dataset.strike_band,
            volatility_scaled=contract_dataset.volatility_scaled,
            volatility_scalar=contract_dataset.volatility_scalar,
            hist_vol=contract_dataset.hist_vol,
            clean_up=True,
        )

        return False, contract
    else:
        if verbose:
            ctx.log(
                f"Contract timespan ({time_span.days} days) is too short. Moving to next contract."
            )
        return True, None


def get_valid_start_date(candidate_start_date: str) -> str:
    """
    Return the next valid NYSE trading day given a candidate date in YYYYMMDD format.

    This function checks whether the provided date falls on a weekend or a NYSE holiday.
    If so, it advances the date forward to the next valid trading day.

    Args:
        candidate_start_date (str): The date to validate, in 'YYYYMMDD' format.

    Returns:
        str: The next valid NYSE trading day in 'YYYYMMDD' format.

    Raises:
        ValueError: If no valid trading day is found within the search buffer.
    """
    nyse = mcal.get_calendar("NYSE")
    date_obj = datetime.strptime(candidate_start_date, "%Y%m%d")

    # Get 10 future trading days (timezone-aware, but we’ll just use .date())
    trading_days = nyse.valid_days(
        start_date=candidate_start_date,
        end_date=(date_obj + timedelta(days=10)).strftime("%Y-%m-%d")
    )

    for d in trading_days:
        if d.date() >= date_obj.date():
            return d.strftime("%Y%m%d")

    raise ValueError(f"No valid trading day found after {candidate_start_date}")


def get_forecasting_loaders(
    train_contract_dataset: ContractDataset,
    val_contract_dataset: ContractDataset,
    test_contract_dataset: ContractDataset,
    seq_len: int,
    pred_len: int,
    tte_tolerance: Tuple[int, int],
    core_feats: List[str] = ["option_returns"],
    tte_feats: Optional[List[str]] = None,
    datetime_feats: Optional[List[str]] = None,
    vol_feats: Optional[List[str]] = None,
    rolling_volatility_range: Optional[List[int]] = None,
    keep_datetime: bool = False,
    target_channels: Optional[List[str]] = None,
    target_type: str = "multistep",
    batch_size: int = 32,
    shuffle: bool = True,
    drop_last: bool = False,
    num_workers: int = 4,
    prefetch_factor: Optional[int] = None,
    pin_memory: bool = torch.cuda.is_available(),
    persistent_workers: bool = True,
    clean_up: bool = False,
    offline: bool = False,
    save_dir: Optional[str] = None,
    verbose: bool = False,
    scaling: bool = False,
    intraday: bool = False,
    dtype: str = "float32",
    modify_contracts: bool = False,
    warning: bool = True,
    dev_mode: bool = False,
) -> Union[
    Tuple[DataLoader, DataLoader, DataLoader, None],
    Tuple[DataLoader, DataLoader, DataLoader, StandardScaler],
]:
    """
    Forms training, validation, and test dataloaders for option contract data.

    Args:
        train_contract_dataset: Contract dataset for training
        val_contract_dataset: Contract dataset for validation
        test_contract_dataset: Contract dataset for testing
        seq_len: Sequence length for input data
        pred_len: Prediction length for forecasting
        tte_tolerance: Tuple of (min, max) time to expiration tolerance in minutes
        core_feats: List of core features to include
        tte_feats: List of time-to-expiration features to include
        datetime_feats: List of datetime features to include
        keep_datetime: Whether to keep the datetime column in the dataset
        target_type: Type of forecasting target. Options: "multistep" (float), "average" (float), or "average_direction" (binary).
        batch_size: Number of samples per batch
        shuffle: Whether to shuffle the data
        drop_last: Whether to drop the last incomplete batch
        num_workers: Number of subprocesses to use for data loading
        prefetch_factor: Number of batches to prefetch
        pin_memory: Whether to pin memory for faster GPU transfer
        clean_up: Whether to clean up the data after use
        offline: Whether to load saved contracts from disk
        save_dir: Directory to save/load processed datasets
        modify_contracts: Whether to modify contracts if they are invalid in get_forecasting_dataset function calls.
        verbose: Whether to print verbose output
        scaling: Whether to normalize the datasets
        intraday: Whether to use intraday data
        target_channels: List of target channels for forecasting
        dtype: Data type for tensors
        warning: Whether to show warnings

    Returns:
        Tuple[DataLoader, DataLoader, DataLoader]: Train, validation, and test data loaders if scaling=False.
        Tuple[DataLoader, DataLoader, DataLoader, StandardScaler]: Train, validation, and test data loaders, and the scaler if scaling=True.
    """

    # Get the combined datasets of contract data for training, validation, and testing
    train_dataset, _ = get_forecasting_dataset(
        contract_dataset=train_contract_dataset,
        tte_tolerance=tte_tolerance,
        seq_len=seq_len,
        pred_len=pred_len,
        core_feats=core_feats,
        tte_feats=tte_feats,
        vol_feats=vol_feats,
        rolling_volatility_range=rolling_volatility_range,
        datetime_feats=datetime_feats,
        keep_datetime=keep_datetime,
        target_type=target_type,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        dtype=dtype,
        modify_contracts=modify_contracts,
        save_dir=save_dir,
        verbose=verbose,
        warning=warning,
        dev_mode=dev_mode,
    )

    val_dataset, _ = get_forecasting_dataset(
        contract_dataset=val_contract_dataset,
        tte_tolerance=tte_tolerance,
        seq_len=seq_len,
        pred_len=pred_len,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        vol_feats=vol_feats,
        rolling_volatility_range=rolling_volatility_range,
        keep_datetime=keep_datetime,
        target_type=target_type,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        dtype=dtype,
        modify_contracts=modify_contracts,
        save_dir=save_dir,
        verbose=verbose,
        warning=warning,
        dev_mode=dev_mode,
    )
    test_dataset, _ = get_forecasting_dataset(
        contract_dataset=test_contract_dataset,
        tte_tolerance=tte_tolerance,
        seq_len=seq_len,
        pred_len=pred_len,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        vol_feats=vol_feats,
        rolling_volatility_range=rolling_volatility_range,
        keep_datetime=keep_datetime,
        target_type=target_type,
        clean_up=clean_up,
        offline=offline,
        intraday=intraday,
        target_channels=target_channels,
        dtype=dtype,
        modify_contracts=modify_contracts,
        save_dir=save_dir,
        verbose=verbose,
        warning=warning,
        dev_mode=dev_mode,
    )

    if scaling:
        train_dataset, val_dataset, test_dataset, scaler = normalize_datasets(
            train_dataset=train_dataset,
            val_dataset=val_dataset,
            test_dataset=test_dataset,
        )
    else:
        scaler = None

    # Create dataloaders for training, validation, and testing
    train_loader = DataLoader(
        dataset=train_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        dataset=val_dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )
    test_loader = DataLoader(
        dataset=test_dataset,
        batch_size=batch_size,
        shuffle=False,
        drop_last=drop_last,
        num_workers=num_workers,
        persistent_workers=persistent_workers,
        prefetch_factor=prefetch_factor,
        pin_memory=pin_memory,
    )

    return train_loader, val_loader, test_loader, scaler


def create_windows(
    df: pd.DataFrame,
    seq_len: int,
    pred_len: int,
    window_stride: int,
    intraday: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Generates rolling windows of data for a given DataFrame. Should be used primarily for scikit-learn models and/or intraday modeling,
    otherwise default to optrade.data.forecasing.get_forecasting_loaders or optrade.data.forecasting.get_forecasting_datasets.

    Args:
        df (pd.DataFrame): DataFrame containing the data.
        seq_len (int): Length of the input sequence.
        pred_len (int): Length of the prediction sequence.
        window_stride (int): Number of steps to move the window forward.
        intraday (bool): Whether the data is intraday or not. If True, the function will first split the data
                         into separate trading days before creating individual windows that cannot crossover
                         between days. Otherwise, the function will create windows that can span multiple days.
    Returns:
        input (np.ndarray): Array of input windows of shape (num_windows, seq_len, num_features) where num_features
                            is the number of columns in the DataFrame (removing datetime but adding returns).
        target (np.ndarray): Array of target windows of shape (num_windows, pred_len, 1).
                             Target contains only returns for the 'option_mid_price'.
    """
    datetime = df["datetime"]
    df_copy = df.copy()

    # Define input features (all columns except datetime)
    feature_columns = [col for col in df_copy.columns if col != "datetime"]
    inputs, targets = [], []

    if intraday:
        # Get all unique days
        days = datetime.dt.date.unique()

        # Iterate through each day independently
        for day in days:
            day_mask = datetime.dt.date == day
            day_data = df_copy.loc[day_mask].copy()

            print(f"Length of data: {len(day_data)}")

            # Since returns will be part of the input features but don't exist for 9:30am
            # we remove the market open (9:30am) of each day
            first_time = day_data["datetime"].iloc[0].time()
            if first_time.hour == 9 and first_time.minute == 30:
                day_data = day_data.iloc[1:].reset_index(drop=True)
                print(f"Day data after removing 9:30am: {day_data.tail()}")

            # Raise an error if the length of the day is less than the sum of seq_len+pred_len
            if len(day_data) < seq_len + pred_len:
                raise ValueError(
                    f"seq_len + pred_len = {seq_len + pred_len} exceeds the length of the day. \
                    Either set intraday=False or reduce seq_len and/or pred_len."
                )

            # Get input features and targets, convert to NumPy arrays
            day_features = day_data[feature_columns].to_numpy()
            day_targets = day_data["option_returns"].to_numpy().reshape(-1, 1)

            # Apply the sliding window technique to obtain windows for inputs and targets
            for i in range(0, len(day_data) - seq_len - pred_len + 1, window_stride):
                inputs.append(day_features[i : i + seq_len])
                targets.append(day_targets[i + seq_len : i + seq_len + pred_len])
    else:
        # Since returns will be part of the input features but don't exist for first market open
        # i.e. 9:30am on the first day, we remove it
        first_time = datetime.iloc[0].time()
        if first_time.hour == 9 and first_time.minute == 30:
            df_copy = df_copy.iloc[1:].reset_index(drop=True)

        # Extract features and targets
        features = df_copy[feature_columns].to_numpy()
        targets_data = df_copy["option_returns"].to_numpy().reshape(-1, 1)

        # Create windows
        for i in range(0, len(df_copy) - seq_len - pred_len + 1, window_stride):
            inputs.append(features[i : i + seq_len])
            targets.append(targets_data[i + seq_len : i + seq_len + pred_len])

    # Convert to numpy arrays
    return np.array(inputs), np.array(targets)


if __name__ == "__main__":
    # Test: get_forecasting_loaders
    root = "AMZN"
    total_start_date = "20230101"
    total_end_date = "20230901"
    right = "C"
    interval_min = 60
    contract_stride = 3
    target_tte = 30
    tte_tolerance = (15, 45)
    moneyness = "ATM"
    volatility_scaled = True
    volatility_scalar = 0.01
    volatility_type = "period"
    strike_band = 0.05

    # TTE features
    tte_feats = ["sqrt", "exp_decay"]

    # Datetime features
    datetime_feats = [
        "sin_minute_of_day",
        "cos_minute_of_day",
        "sin_hour_of_week",
        "cos_hour_of_week",
    ]

    # Select features
    core_feats = [
        "log_option_returns",
        "log_stock_returns",
        "option_returns",
        "stock_returns",
        "distance_to_strike",
        "moneyness",
        "option_lob_imbalance",
        "option_quote_spread",
        "stock_lob_imbalance",
        "stock_quote_spread",
        "option_mid_price",
        "option_bid_size",
        "option_bid",
        "option_ask_size",
        "option_close",
        "option_volume",
        "option_count",
        "stock_mid_price",
        "stock_bid_size",
        "stock_bid",
        "stock_ask_size",
        "stock_ask",
        "stock_volume",
        "stock_count",
    ]

    # Testing: get_loaders
    from optrade.data.contracts import get_contract_datasets

    train_cd, val_cd, test_cd = get_contract_datasets(
        root=root,
        start_date=total_start_date,
        end_date=total_end_date,
        contract_stride=contract_stride,
        interval_min=interval_min,
        right=right,
        target_tte=target_tte,
        tte_tolerance=tte_tolerance,
        moneyness=moneyness,
        strike_band=strike_band,
        volatility_type=volatility_type,
        volatility_scaled=volatility_scaled,
        volatility_scalar=volatility_scalar,
        verbose=True,
        train_split=0.5,
        val_split=0.25,
        dev_mode=True,
    )

    output = get_forecasting_loaders(
        train_contract_dataset=train_cd,
        val_contract_dataset=val_cd,
        test_contract_dataset=test_cd,
        tte_tolerance=tte_tolerance,
        keep_datetime=True,
        target_type="average",
        seq_len=100,
        pred_len=10,
        core_feats=core_feats,
        tte_feats=tte_feats,
        datetime_feats=datetime_feats,
        batch_size=32,
        clean_up=False,
        offline=True,
        save_dir=None,
        verbose=True,
        scaling=True,
        dev_mode=True,
    )
    train_loader, val_loader, test_loader = output[0:3]

    print(f"Num train examples: {len(train_loader.dataset)}")
    print(f"Num val examples: {len(val_loader.dataset)}")
    print(f"Num test examples: {len(test_loader.dataset)}")

    for batch in train_loader:
        x, y, x_dt, y_dt = batch
        print(f"x ({x.dtype}) shape: {x.shape}. y ({y.dtype}) shape: {y.shape}")

        print(f"x_dt (before): {x_dt.shape}. y_dt (before): {y_dt.shape}")
        x_dt = tensor_to_datetime(timestamp_tensor=x_dt, batch_mode=True)
        y_dt = tensor_to_datetime(timestamp_tensor=y_dt, batch_mode=True)

        print(f"x_dt: {x_dt}. y_dt: {y_dt}")
        break

    # # Testing: create_windows
    # from optrade.data.features import transform_features
    # from optrade.data.contracts import Contract
    # from rich.console import Console

    # console = Console()

    # contract = Contract.find_optimal(
    #     root="AAPL",
    #     start_date="20241107",
    #     volatility_scaled=False,
    #     strike_band=0.05,
    #     moneyness="OTM",
    #     interval_min=1,
    #     right="C",
    #     target_tte=30,
    #     tte_tolerance=(25, 35),
    # )

    # df = contract.load_data(clean_up=True, offline=False, warning=True)

    # # TTE features
    # tte_feats = ["sqrt", "exp_decay"]

    # # Datetime features
    # datetime_feats = [
    #     "sin_minute_of_day",
    #     "cos_minute_of_day",
    #     "sin_hour_of_week",
    #     "cos_hour_of_week",
    # ]

    # # Select features
    # core_feats = [
    #     "option_returns",
    #     "stock_returns",
    #     "distance_to_strike",
    #     "moneyness",
    #     "option_lob_imbalance",
    #     "option_quote_spread",
    #     "stock_lob_imbalance",
    #     "stock_quote_spread",
    #     "option_mid_price",
    #     "option_bid_size",
    #     "option_bid",
    #     "option_ask_size",
    #     "option_close",
    #     "option_volume",
    #     "option_count",
    #     "stock_mid_price",
    #     "stock_bid_size",
    #     "stock_bid",
    #     "stock_ask_size",
    #     "stock_ask",
    #     "stock_volume",
    #     "stock_count",
    # ]

    # df = transform_features(
    #     df=df,
    #     core_feats=core_feats,
    #     tte_feats=tte_feats,
    #     datetime_feats=datetime_feats,
    #     strike=contract.strike,
    #     exp=contract.exp,
    #     keep_datetime=True,
    # )
    # print(df.columns)

    # x, y = create_windows(
    #     df=df, seq_len=30, pred_len=6, window_stride=1, intraday=False
    # )

    # print(x.shape, y.shape)
